#!/usr/bin/env python
"""General functions for curating a model

This module provides functionalities for curating models, including special functions for CarveMe models.

Since CarveMe version 1.5.1, the draft models from CarveMe contain pieces of information that are not correctly added to the
annotations. To address this, this module includes the following functionalities:

    - Add URIs from the entity IDs to the annotation field for metabolites & reactions
    - Transfer URIs from the notes field to the annotations for metabolites & reactions
    - Add URIs from the GeneProduct IDs to the annotations

The functionalities for CarveMe models, along with some of the following further functionalities, are gathered in the
main function :py:func:`~refinegems.curation.curate.polish_model`.

Further functionalities:

        - Setting boundary condition & constant for metabolites & reactions
        - Unit handling to add units & UnitDefinitions & to set units for parameters
        - Addition of default settings for compartments & metabolites
        - Addition of URIs to GeneProducts

            - via a mapping from model IDs to valid database IDs
            - via the KEGG API

        - Changing the CURIE pattern/CVTerm qualifier & qualifier type
        - Directionality control
"""

__author__ = "Famke Baeuerle and Carolin Brune and Gwendolyn O. Döbel"

################################################################################
# requirements
################################################################################

import cobra
import logging
import pandas as pd
import re
import warnings

from bioservices.kegg import KEGG
from cobra.io.sbml import _f_specie, _f_reaction
from libsbml import Model as libModel
from libsbml import GeneProduct, Species, ListOfSpecies, ListOfReactions, UnitDefinition
from pathlib import Path
from tqdm.auto import tqdm
from typing import Literal, Union

from .miriam import polish_annotations, change_all_qualifiers

from ..utility.cvterms import (
    add_cv_term_genes,
    add_cv_term_metabolites,
    add_cv_term_reactions,
    DB2PREFIX_METABS,
    DB2PREFIX_REACS,
    get_id_from_cv_term,
)
from ..utility.entities import get_gpid_mapping, create_fba_units, print_UnitDefinitions
from ..utility.io import load_a_table_from_database
from ..utility.util import DB2REGEX, test_biomass_presence

################################################################################
# variables
################################################################################

NH_PATTERN = re.compile(r"nh[3-4]")  #: :meta:

################################################################################
# functions
################################################################################

# sychronise annotations
# ----------------------


def update_annotations_from_others(model: libModel) -> libModel:
    """Synchronizes metabolite annotations for core, periplasm and extracelullar

    Args:
        - model (libModel):
            Model loaded with libSBML

    Returns:
        libModel:
            Modified model with synchronized annotations
    """
    for metab in model.getListOfSpecies():
        base = metab.getId()[:-2]
        for comp in ["_c", "_e", "_p"]:
            other_metab = model.getSpecies(base + comp)
            if other_metab is not None:
                if not other_metab.isSetMetaId():
                    other_metab.setMetaId("meta_" + other_metab.getId())
                for db_id, code in DB2PREFIX_METABS.items():
                    id = get_id_from_cv_term(metab, code)
                    for entry in id:
                        if entry is not None:
                            add_cv_term_metabolites(entry, db_id, other_metab)
    return model


def extend_gp_annots_via_mapping_table(
    model: libModel,
    mapping_tbl_file: Union[str,Path] = None,
    gff_paths: list[str] = None,
    email: str = None,
    contains_locus_tags: bool = False,
    lab_strain: bool = False,
    outpath: str = None,
) -> libModel:
    """
    | Extend GenePoduct annotations via mapping table.
    | If no mapping table is provided, a mapping table will be generated.

    Args:
        - model (libModel):
            Model loaded with libSBML
        - mapping_tbl_file (str|Path, optional):
            Path to a file containing a mapping table with columns ``model_id | X...`` where X can be ``REFSEQ``,
            ``NCBI``, ``locus_tag`` or ``UNCLASSIFIED``.
            The table can contain all of the ``X`` columns or at least one of them.
            Defaults to None.
        - gff_paths (list[str], optional):
            Path(s) to GFF file(s). Allowed GFF formats are: RefSeq, NCBI and Prokka.
            This is only used when mapping_tbl_file == None.
            Defaults to None.
        - email (str, optional):
            E-mail for NCBI queries.
            This is only used when mapping_tbl_file == None.
            Defaults to None.
        - contains_locus_tags (bool, optional):
            Specifies if provided model has locus tags within the label tag if set to True.
            This is only used when mapping_tbl_file == None.
            Defaults to False.
        - lab_strain (bool, optional):
            Specifies if a strain from no database was provided and thus has only homolog mappings if set to True.
            Defaults to False.
        - outpath (str, optional):
            Output path for location where the generated mapping table should be written to.
            This is only used when mapping_tbl_file == None.
            Defaults to None.

    Returns:
        libModel:
            Modified model with extended annotations for the GeneProducts
    """

    # 1. Get mapping
    # If no mapping table provided, get via function
    print("Get mapping information...")
    if not mapping_tbl_file:
        mapping_table = get_gpid_mapping(
            model, gff_paths, email, contains_locus_tags, outpath
        )
    else:  # Otherwise read in table from file
        mapping_tbl_file = mapping_tbl_file if isinstance(mapping_tbl_file, str) else str(mapping_tbl_file)
        mapping_table = pd.read_csv(mapping_tbl_file)

    # Drop all rows without model_id entries
    mapping_table.dropna(subset="model_id", inplace=True)
    # Use model_id as index
    mapping_table.set_index("model_id", inplace=True)
    # Replace all NaN values with empty string
    mapping_table.fillna("", inplace=True)

    # 2. Use table to fill in information in model
    # Get gene list
    gene_list = model.getPlugin("fbc").getListOfGeneProducts()

    print("Extending GeneProduct information...")
    for gene in tqdm(gene_list):
        # Get row of mapping table for current model_id
        gp_infos = mapping_table.loc[gene.getId(), :]

        # Add infos to current GeneProduct
        if gp_infos["name"]:
            gene.setName(gp_infos["name"])
        if gp_infos["locus_tag"]:
            gene.setLabel(gp_infos["locus_tag"])
        if ("REFSEQ" in gp_infos.index.to_list()) and gp_infos["REFSEQ"]:
            add_cv_term_genes(gp_infos["REFSEQ"], "REFSEQ", gene, lab_strain)
        if ("NCBI" in gp_infos.index.to_list()) and gp_infos["NCBI"]:
            add_cv_term_genes(gp_infos["NCBI"], "NCBI", gene, lab_strain)

    return model


def extend_gp_annots_via_KEGG(gene_list: list[GeneProduct], kegg_organism_id: str):
    """Adds KEGG gene & UniProt identifiers to the GeneProduct annotations

    Args:
        gene_list (list[GeneProduct]):
            libSBML ListOfGenes
        kegg_organism_id (str):
            Organism identifier in the KEGG database
    """
    k = KEGG()
    mapping_kegg_uniprot = k.conv("uniprot", kegg_organism_id)
    no_valid_kegg = []

    for gp in tqdm(gene_list):

        if gp.getId() != "G_spontaneous":
            kegg_gene_id = f"{kegg_organism_id}:{gp.getLabel()}"

            try:
                uniprot_id = mapping_kegg_uniprot[kegg_gene_id]

                add_cv_term_genes(kegg_gene_id, "KEGG", gp)
                add_cv_term_genes(uniprot_id.split(r"up:")[1], "UNIPROT", gp)

            except KeyError:
                no_valid_kegg.append(gp.getLabel())

    if no_valid_kegg:
        logging.info(
            f"The following {len(no_valid_kegg)} locus tags form no valid KEGG Gene ID: {no_valid_kegg} with the provided KEGG Organism ID: {kegg_organism_id}."
        )


def extend_metab_reac_annots_via_id(
    entity_list: Union[ListOfSpecies, ListOfReactions], id_db: str
) -> None:
    """Extends metabolite or reaction annotations by extracting the core ID from the entity ID and adding this ID as
    valid annotation if possible

    Args:
        - entity_list (Union[ListOfSpecies, ListOfReactions]):
            List of entities, either species (metabolites, ListOfSpecies) or reactions (ListOfReactions)
        - id_db (str):
            The database prefix to validate IDs against. Must correspond to a valid
            prefix in the Bioregistry

    Raises:
        - TypeError: Unsupported type for entity_list
    """
    # Initialise id_db correctly
    id_db = id_db.upper()

    # Set-up case-dependent variables
    match entity_list:
        case ListOfSpecies():
            bigg_db = "bigg_metabolites"
            bigg_id_type = "universal_bigg_id"
            add_cv_term = add_cv_term_metabolites
            id_handler = _f_specie
            db2prefix = DB2PREFIX_METABS
        case ListOfReactions():
            bigg_db = "bigg_reactions"
            bigg_id_type = "id"
            add_cv_term = add_cv_term_reactions
            id_handler = _f_reaction
            db2prefix = DB2PREFIX_REACS
        case _:
            raise TypeError(
                f"Unsupported type for entity_list {type(entity_list)}. Must be ListOfSpecies or ListOfReactions."
            )

    # Set-up default variables
    try:
        id_db_prefix = db2prefix[id_db]
    except KeyError:
        print(
            f"""
KeyError: with id_db=\'{id_db}\'
id_db must be one of the valid database names: {db2prefix.keys()}
If your id_db is not part of the list, please contact the developers.
              """
        )
        return
    try:
        db_pattern = DB2REGEX[id_db_prefix]
    except KeyError:
        print(
            f"KeyError: id_db_prefix = {id_db_prefix} must be one of the valid prefixes in https://bioregistry.io/."
        )
        return

    # Get BiGG IDs for VMH ID == BiGG ID validation
    if "vmh" in id_db_prefix.lower():
        bigg_ids = load_a_table_from_database(f"SELECT bigg_id FROM {bigg_db}")
        bigg_ids = set(bigg_ids[bigg_id_type].tolist())

    # Get ID from entity & add annotation if valid database ID
    for entity in entity_list:

        # Get ID
        current_id = entity.getId()

        # Use current_id as metaid if no metaid is present
        if not entity.isSetMetaId():
            entity.setMetaId(f"meta_{current_id}")

        if re.fullmatch(db_pattern, current_id, re.IGNORECASE):

            # Remove prefix
            current_id = id_handler(current_id)

            # Unset annotations if no CV terms exist
            if entity.getNumCVTerms() == 0:
                entity.unsetAnnotation()

            # Get ID for annotation
            if isinstance(entity, Species):  # Remove compartment suffix
                id_for_anno = re.sub(
                    f"(_|\[){entity.getCompartment()}\]?$", "", current_id
                )
            else:
                id_for_anno = current_id

            # Add ID as URI to annotation
            add_cv_term(id_for_anno, id_db, entity)

            # Add BiGG ID to annotation, additionally, if VMH and valid BiGG ID
            if ("vmh" in id_db_prefix.lower()) and (id_for_anno in bigg_ids):
                add_cv_term(id_for_anno, id_db, entity)


def extend_metab_reac_annots_via_notes(
    entity_list: Union[ListOfSpecies, ListOfReactions],
) -> None:
    """Extends metabolite or reaction annotations by extracting valid URIs from the notes section of the provided
    entities and cleans up the notes by removing processed elements

    Args:
        - entity_list (Union[ListOfSpecies, ListOfReactions]):
            List of entities, either species (metabolites, ListOfSpecies) or reactions (ListOfReactions)
    Raises:
        - TypeError: Unsupported type for entity_list
    """

    # Set-up case-dependent variables
    match entity_list:
        case ListOfSpecies():
            db2prefix = DB2PREFIX_METABS
            add_cv_term = add_cv_term_metabolites
        case ListOfReactions():
            db2prefix = DB2PREFIX_REACS
            add_cv_term = add_cv_term_reactions
        case _:
            raise TypeError(
                f"Unsupported type for entity_list {type(entity_list)}. Must be ListOfSpecies or ListOfReactions."
            )

    # Get notes & add annotation from notes to CVTerms if valid URI
    for entity in entity_list:

        if not entity.isSetMetaId():
            entity.setMetaId("meta_" + entity.getId())

        notes_list = []
        elem_used = []
        notes_string = entity.getNotesString().split("\n")
        for elem in notes_string:
            for id_db in db2prefix.keys():
                if "<p>" + id_db in elem.upper():
                    elem_used.append(elem)
                    # @DEBUG print(elem.strip()[:-4].split(r': ')[1])
                    fill_in = re.split(r":\s*", elem.strip()[:-4])[1]
                    if (";") in fill_in:
                        if not re.search(r"inchi", id_db, re.IGNORECASE):
                            entries = fill_in.split(r";")
                            for entry in entries:
                                if not re.fullmatch(
                                    r"^nan$", entry.strip(), re.IGNORECASE
                                ):
                                    add_cv_term(entry.strip(), id_db, entity)
                    else:
                        if not re.fullmatch(r"^nan$", fill_in, re.IGNORECASE):
                            add_cv_term(fill_in.strip(), id_db, entity)

        for elem in notes_string:
            if elem not in elem_used and elem not in notes_list:
                notes_list.append(elem)

        # Adding new, shortened notes
        new_notes = " ".join([str(elem) + "\n" for elem in notes_list])
        entity.unsetNotes()
        entity.setNotes(new_notes)
        # @DEBUG print(species.getAnnotationString())


# correct basic model set-up
# ---------------------------


def polish_model_units(model: libModel) -> None:
    """Replaces the list of unit definitions with the unit definitions needed for FBA:

       - mmol per gDW per h
       - mmol per gDW
       - hour (h)
       - femto litre (fL)

    Args:
        - model (libModel):
            Model loaded with libSBML
    """

    # Get FBA unit definitions per refineGEMs definition
    fba_unit_defs = create_fba_units(model)

    # Get model unit definitions
    model_unit_defs = model.getListOfUnitDefinitions().clone()

    # If list of unit definitions is not empty, replace all units with the defined FBA units
    # & Print the non-FBA unit definitions
    if model_unit_defs:

        # List to collect all non-FBA unit definitions
        removed_unit_defs = []

        # Check if model unit definitions fit to the fba unit definitions
        for model_ud in model_unit_defs:
            for fba_ud in fba_unit_defs:

                # In case of identical unit definitions, remove unit definition from fba unit def list
                if not UnitDefinition.areIdentical(fba_ud, model_ud):
                    removed_unit_defs.append(model_ud)

        # Remove all model unit definitions
        model.getListOfUnitDefinitions().clear(doDelete=True)

        # Only print list if UnitDefinitions were removed
        if len(removed_unit_defs) == 0:
            logging.warning(
                """
            The following UnitDefinition objects were removed. 
            The reasoning is that
            \t(a) these UnitDefinitions are not contained in the UnitDefinition list of this program and
            \t(b) the UnitDefinitions defined within this program are handled as ground truth.
            Thus, the following UnitDefinitions are not seen as relevant for the model.
            """
            )
            print_UnitDefinitions(removed_unit_defs)

    # Add all defined FBA units to the model
    for unit_def in fba_unit_defs:
        model.getListOfUnitDefinitions().append(unit_def)


def set_model_default_units(model: libModel):
    """Sets default units of model

    Args:
        - model (libModel):
            Model loaded with libSBML
    """
    for unit in model.getListOfUnitDefinitions():

        unit_id = unit.getId()

        if re.fullmatch(r"mmol_per_gDW", unit_id, re.IGNORECASE):

            if not (model.isSetExtentUnits() and model.getExtentUnits() == unit_id):
                model.setExtentUnits(unit_id)

            if not (
                model.isSetSubstanceUnits() and model.getSubstanceUnits() == unit_id
            ):
                model.setSubstanceUnits(unit_id)

        if not (
            model.isSetTimeUnits() and model.getTimeUnits() == unit_id
        ) and re.fullmatch(r"hr?", unit_id, re.IGNORECASE):
            model.setTimeUnits(unit_id)

        if not (
            model.isSetVolumeUnits() and model.getVolumeUnits() == unit_id
        ) and re.fullmatch(r"fL", unit_id, re.IGNORECASE):
            model.setVolumeUnits(unit_id)


def set_units_of_parameters(model: libModel):
    """Sets units of parameters in model

    Args:
        - model (libModel):
            Model loaded with libSBML
    """
    for (
        param
    ) in (
        model.getListOfParameters()
    ):  # needs to be added to list of unit definitions aswell
        if any(
            (
                unit_id := re.fullmatch(
                    r"mmol_per_gDW_per_hr?", unit.getId(), re.IGNORECASE
                )
            )
            for unit in model.getListOfUnitDefinitions()
        ):
            if not (param.isSetUnits() and param.getUnits() == unit_id.group(0)):
                param.setUnits(unit_id.group(0))


def add_compartment_structure_specs(model: libModel):
    """| Adds the required specifications for the compartment structure
    | if not set (size & spatial dimension)

    Args:
        - model (libModel):
            Model loaded with libSBML
    """
    for compartment in model.getListOfCompartments():

        if not compartment.isSetSize():
            compartment.setSize(float("NaN"))

        if not compartment.isSetSpatialDimensions():
            compartment.setSpatialDimensions(3)

        if any(
            (unit_id := re.fullmatch(r"fL", unit.getId(), re.IGNORECASE))
            for unit in model.getListOfUnitDefinitions()
        ):
            if not (
                compartment.isSetUnits() and compartment.getUnits() == unit_id.group(0)
            ):
                compartment.setUnits(unit_id.group(0))


def set_initial_amount_metabs(model: libModel):
    """Sets initial amount to all metabolites if not already set or if initial concentration is not set

    Args:
        - model (libModel):
            Model loaded with libSBML
    """
    for species in model.getListOfSpecies():

        if not (species.isSetInitialAmount() or species.isSetInitialConcentration()):
            species.setInitialAmount(float("NaN"))


def polish_entity_conditions(entity_list: Union[ListOfSpecies, ListOfReactions]):
    """Sets boundary condition and constant if not set for an entity

    Args:
        - entity_list (Union[ListOfSpecies, ListOfReactions]):
            libSBML ListOfSpecies or ListOfReactions
    """
    match entity_list:
        case ListOfSpecies():
            for entity in entity_list:
                if not entity.getBoundaryCondition():
                    entity.setBoundaryCondition(False)
                if not entity.getConstant():
                    entity.setConstant(False)
        case ListOfReactions():
            pass
        case _:
            logging.warning(
                f"Unsupported type for entity_list {type(entity_list)}. Must be ListOfSpecies or ListOfReactions."
            )


# duplicates
# ----------


def resolve_duplicate_reactions(
    model: cobra.Model, based_on: str = "reaction", remove_reac: bool = True
) -> cobra.Model:
    """Resolve and remove duplicate reaction based on their reaction equation
    and matching database identifiers. Only if all match or a comparison with nan occurs will one of
    the reactions be removed.

    Args:
        - model (cobra.Model):
            A model loaded with COBRApy.
        - based_on (str, optional):
            Label to base the resolvement process on .
            Can be 'reaction' or any other annotation label.
            Defaults to 'reaction'.
        - remove_reac (bool, optional):
            When True, combines and remove duplicates.
            Otherwise only reports the findings.
            Defaults to True.

    Returns:
        cobra.Model:
            The model.
    """

    # get annotation and compartment information
    anno_reac = []
    for r in model.reactions:
        anno_reac.append(
            {"id": r.id, "compartment": str(r.compartments), "reaction": r.reaction}
            | r.annotation
        )
    df_reac = pd.DataFrame.from_dict(anno_reac)

    # check if based_on is valid
    if not based_on in df_reac.columns.tolist():
        warnings.warn(
            f"Warning: Annotation column {based_on} does not exists. Search for duplicates will be skipped."
        )
        return model

    # set basic parameters
    skip_cols = ["id", "compartment", "bigg.reaction", "reaction", based_on]
    colnames = df_reac.columns.tolist()

    for c in df_reac.groupby("compartment"):
        # note: using groupby drops nans
        for mnx in c[1].groupby(based_on):
            # find possible duplicates
            dupl = True
            annotations = {}
            if len(mnx[1]) > 1:
                # check annotations
                for col in [_ for _ in colnames if not _ in skip_cols]:
                    if len(mnx[1][col].dropna().value_counts()) < 2:
                        annotations[col] = (
                            mnx[1][col].dropna().explode().unique().tolist()
                        )
                    else:
                        dupl = False
                        break

                # if duplicate found
                if dupl:
                    if remove_reac:

                        # choose reaction to keep
                        keep_reac = model.reactions.get_by_id(mnx[1]["id"].tolist()[0])

                        # resolve annotations
                        for key, value in annotations.items():
                            if len(value) > 0 and not key in keep_reac.annotation:
                                keep_reac.annotation[key] = value

                        # combine gene reaction rules
                        for r_id in mnx[1]["id"].tolist()[1:]:
                            gpr_to_add = model.reactions.get_by_id(
                                r_id
                            ).gene_reaction_rule
                            if gpr_to_add and gpr_to_add != "":
                                if (
                                    keep_reac.gene_reaction_rule
                                    and keep_reac.gene_reaction_rule != ""
                                ):  # add two existing rules
                                    keep_reac.gene_reaction_rule = (
                                        keep_reac.gene_reaction_rule
                                        + " or "
                                        + gpr_to_add
                                    )
                                else:
                                    keep_reac.gene_reaction_rule = (
                                        gpr_to_add  # add the one existing the other
                                    )
                            else:
                                pass  # nothing to add
                            model.reactions.get_by_id(r_id).remove_from_model()
                            print(
                                f"\tDuplicate reaction {r_id} found. Combined to {keep_reac.id} and deleted."
                            )
                    else:
                        print(
                            f'\tDuplicate reactions {", ".join(mnx[1]["id"].tolist())} found.'
                        )

    return model


def resolve_duplicate_metabolites(
    model: cobra.Model, based_on: str = "metanetx.chemical", replace: bool = True
) -> cobra.Model:
    """Resolve duplicate metabolites in a model. Metabolites are considered
    duplicate if they share the same annotations (same or nan).

    .. note::
        Depending on the starting database, the results might differ.

    Args:
        - model (cobra.Model):
            The model loaded with COBRApy.
        - based_on (str, optional):
            Label to base the resolvement process on .
            Can be any annotation label.
            Defaults to 'metanetx.chemical'.
        - replace (bool, optional):
            Either report the duplicates (False)
            or replace them with one (True). Defaults to True.

    Returns:
        cobra.Model:
            The model.
    """

    # get annotation and compartment information
    anno_meta = []
    for m in model.metabolites:
        anno_meta.append({"id": m.id, "compartment": m.compartment} | m.annotation)
    df_meta = pd.DataFrame.from_dict(anno_meta)

    # check if based_on is valid
    if not based_on in df_meta.columns.tolist():
        warnings.warn(
            f"Warning: Annotation {based_on} not found. Search for metabolite duplicates skipped."
        )
        return model

    # set basic parameters
    skip_cols = ["id", "compartment", "bigg.metabolite", based_on]
    colnames = df_meta.columns.tolist()
    # get objective function
    bof_list = test_biomass_presence(model)
    if len(bof_list) == 1:
        objective_function = bof_list[0]
    elif len(bof_list) > 1:
        mes = f"Multiple BOFs detected. Will be using {bof_list[0]}"
        warnings.warn(mes, category=UserWarning)
        objective_function = bof_list[0]
    else:
        mes = "No BOF detected. Might lead to problems during duplicate removal."
        warnings.warn(mes, category=UserWarning)

    for c in df_meta.groupby("compartment"):
        # note: using groupby drops nans
        # note: mnx as starting point was choosen, as it seems to have the most annotations (easy to get)
        c[1][based_on] = c[1][based_on].apply(
            lambda x: tuple(x) if isinstance(x, list) else x
        )
        for mnx in c[1].groupby(based_on):

            # find possible duplicates
            dupl = True
            annotations = {}
            if len(mnx[1]) > 1:
                for col in [_ for _ in colnames if not _ in skip_cols]:
                    # check annotation, if current group truly consists of duplicates
                    if len(mnx[1][col].dropna().value_counts()) < 2:
                        annotations[col] = (
                            mnx[1][col].dropna().explode().unique().tolist()
                        )
                    else:
                        dupl = False
                        break
            # no duplicates = check next
            else:
                continue

            # if duplicate is found:
            if dupl:
                # either replace ...
                if replace:
                    # choose metabolite to keep
                    keep_meta = model.metabolites.get_by_id(mnx[1]["id"].tolist()[0])
                    # resolve annotations
                    for key, value in annotations.items():
                        if len(value) > 0 and not key in keep_meta.annotation:
                            keep_meta.annotations[key] = value
                    # note: charge and formula should be in valid range and be corrected by MCC (if needed)

                    # replace duplicates with the metabolite to be kept
                    #     to ensure consistency, only delete duplicate metabolites, which
                    #     do NOT share ANY reactions

                    # retrieve reactions for metabolites set for keeping
                    keep_reac = [
                        _.id
                        for _ in model.metabolites.get_by_id(keep_meta.id).reactions
                    ]
                    # iterate over metabolites set for deletion
                    for del_meta_id in mnx[1]["id"].tolist()[1:]:
                        # retrieve reaction for metabolites set for deletion
                        del_reac = [
                            _.id
                            for _ in model.metabolites.get_by_id(del_meta_id).reactions
                        ]
                        # get intersection of reactions (keep + del)
                        reac_intersec = list(set(keep_reac) & set(del_reac))

                        # if intersection empty, metabolite is with a high probability indeed a duplicate
                        # Special case: NH3 / NH4
                        #    intersection does not have to be emtpy, know 'problem' caused by CarveMe
                        if len(reac_intersec) == 0 or all(
                            [
                                re.search(NH_PATTERN, _)
                                for _ in [keep_meta.id, del_meta_id]
                            ]
                        ):
                            # automated deletion is only advisable, if consistency can be retained
                            perform_deletion = True
                            with model as model_del:

                                # if the special case is detected ...
                                if all(
                                    [
                                        re.search(NH_PATTERN, _)
                                        for _ in [keep_meta.id, del_meta_id]
                                    ]
                                ):
                                    print(
                                        f"\tSpecial case -Duplicate NH4/NH3- detected.\n\tTrying to solve by additionally removing reactions containing both metabolites."
                                    )
                                    # ... remove reactions with nh3 and nh4 both present
                                    for del_reac_id in reac_intersec:
                                        # if objective_function is part of the set
                                        # automated deletion is (currently) not possible
                                        if del_reac_id == objective_function:
                                            perform_deletion = False
                                            break
                                        model_del.reactions.get_by_id(
                                            del_reac_id
                                        ).remove_from_model()
                                    # set the metabolites to be deleted to be the one NOT in the objective functions
                                    # to avoid inconsistencies
                                    if del_meta_id in [
                                        _.id
                                        for _ in model_del.reactions.get_by_id(
                                            objective_function
                                        ).metabolites
                                    ]:
                                        temp = del_meta_id
                                        del_meta_id = keep_meta.id
                                        keep_meta = model_del.metabolites.get_by_id(
                                            temp
                                        )

                                # try replacing metabolite with the kept duplicate ...
                                for reac in model_del.metabolites.get_by_id(
                                    del_meta_id
                                ).reactions:

                                    reac.reaction = reac.reaction.replace(
                                        del_meta_id, keep_meta.id
                                    )

                                    # skip if objective function is found
                                    if reac.id == objective_function:
                                        continue

                                    # check if consistency is still intact
                                    balance_test = reac.check_mass_balance()
                                    if not reac.boundary and len(balance_test) > 0:
                                        # try fixing H-balance
                                        if "H" in balance_test.keys():
                                            # ..............................
                                            # TODO:
                                            #    get H according to compartment
                                            #    current implementation relies heavily
                                            #    on 'correct' use input: compartment should have format C_c or c (C_p, p, C_e, e etc.)
                                            # ..............................
                                            reac_comp = reac.compartments.pop()[-1]
                                            if reac_comp == "c":
                                                reac.subtract_metabolites(
                                                    {"h_c": balance_test["H"]}
                                                )
                                            elif reac_comp == "p":
                                                reac.subtract_metabolites(
                                                    {"h_p": balance_test["H"]}
                                                )
                                            elif reac_comp == "e":
                                                reac.subtract_metabolites(
                                                    {"h_e": balance_test["H"]}
                                                )
                                            else:
                                                perform_deletion = False
                                                break
                                        # ..............................
                                        # TODO:
                                        #    fix other possible problems
                                        # ..............................

                                        # finally, check balance again (continue only if fixed, else break)
                                        if len(reac.check_mass_balance()) > 0:
                                            perform_deletion = False
                                            break

                                    else:
                                        continue

                                # if not problems are found, duplicate is removed
                                if perform_deletion:
                                    model = model_del.copy()
                                    print(
                                        f"\tDuplicate metabolite {del_meta_id} found. Replaced with {keep_meta.id}."
                                    )
                                # if problems are not solvable, duplicate is kept and only reported
                                else:
                                    print(
                                        f"\tDuplicate metabolite {del_meta_id} found (duplicate to {keep_meta.id} based on annotation).\n\t\tAutomated deletion not possible due to problems with consistency."
                                    )

                        # else, metabolite is kept
                        #       since it might be an isomer, elongation, or other explanation
                        #       for the same annotation
                        else:
                            print(
                                f"\tDuplicate metabolite {del_meta_id} found (duplicate to {keep_meta.id} based on annotation).\n\t\tKept, as reaction containing both metabolites was found."
                            )

                # ... or only report duplicates
                else:
                    print(
                        f'\tDuplicate metabolite(s) {", ".join(mnx[1]["id"].tolist())} found.'
                    )

    return model


def resolve_duplicates(
    model: cobra.Model,
    check_reac: bool = True,
    check_meta: Literal["default", "exhaustive", "skip"] = "default",
    replace_dupl_meta: bool = True,
    remove_unused_meta: bool = False,
    remove_dupl_reac: bool = True,
) -> cobra.Model:
    """Resolve and remove (optional) duplicate metabolites and reactions in the model.

    Args:
        - model (cobra.Model):
            The model loaded with COBRApy.
        - check_reac (bool, optional):
            Whether to check reactions for duplicates.
            Defaults to True.
        - check_meta (Literal['default','exhaustive','skip'], optional):
            Whether to check for duplicate metabolites.
            Defaults to 'default'.
        - replace_dupl_meta (bool, optional):
            Option to replace/remove duplicate metabolites.
            Defaults to True.
        - remove_unused_meta (bool, optional):
            Option to remove unused metabolites.
            Defaults to False.
        - remove_dupl_reac (bool, optional):
            Option to combine/remove duplicate reactions.
            Defaults to True.

    Returns:
        cobra.Model:
            The (edited) model.
    """

    # resolve duplicate metabolites
    if check_meta == "default":
        # resolve duplicates starting with the metanetx.chemical database identifiers
        model = resolve_duplicate_metabolites(model, replace=replace_dupl_meta)
    elif check_meta == "exhaustive":
        # resolve duplicates by starting at every database identifer one after another
        # note: bigg and sbo are skipped as sbo gives not much information and bigg is
        #       usually the one that differs (naming issue)
        anno_types = set()
        # get all database annotation types present in the model
        for m in model.metabolites:
            anno_types = anno_types | set(m.annotation.keys())
        for colname in [_ for _ in anno_types if not _ in ["bigg.metabolite", "sbo"]]:
            model = resolve_duplicate_metabolites(
                model, colname, replace=replace_dupl_meta
            )
    elif check_meta == "skip":
        print("\tSkip check for duplicate metabolites.")
    else:
        warnings.warn(
            f"Warning: Unknown option for metabolites duplicate checking {check_meta}. Search for metabolite duplicates skipped."
        )

    # remove now unused metabolites
    if remove_unused_meta:
        model, removed = cobra.manipulation.delete.prune_unused_metabolites(model)
        print(
            f'\tThe following metabolites () have been removed: {", ".join([x.id for x in removed])}'
        )

    # resolve duplicate reactions
    if check_reac:
        model = resolve_duplicate_reactions(
            model, based_on="reaction", remove_reac=remove_dupl_reac
        )

    return model


# Directionality Control
# ----------------------


def check_direction(model: cobra.Model, data: Union[pd.DataFrame, str]) -> cobra.Model:
    """Check the direction of reactions by searching for matching MetaCyc,
    KEGG and MetaNetX IDs as well as EC number in a downloaded BioCyc (MetaCyc)
    database table or dataFrame (need to contain at least the following columns:
    Reactions (MetaCyc ID),EC-Number,KEGG reaction,METANETX,Reaction-Direction.

    Args:
        model (cobra.Model):
            The model loaded with COBRApy.
        data (pd.DataFrame | str):
            Either a pandas DataFrame or a path to a CSV file
            containing the BioCyc smart table.

    Raises:
        - TypeError: Unknown data type for parameter data

    Returns:
        cobra.Model:
            The edited model.
    """

    match data:
        # already a DataFrame
        case pd.DataFrame():
            pass
        case str():
            # load from a table
            data = pd.read_csv(data, sep="\t")
            # rewrite the columns into a better comparable/searchable format
            data["KEGG reaction"] = data["KEGG reaction"].str.extract(r".*>(R\d*)<.*")
            data["METANETX"] = data["METANETX"].str.extract(r".*>(MNXR\d*)<.*")
            data["EC-Number"] = data["EC-Number"].str.extract(r"EC-(.*)")
        case _:
            mes = f"Unknown data type for parameter data: {type(data)}"
            raise TypeError(mes)

    # check direction
    # --------------------
    for r in model.reactions:

        direction = None
        # easy case: metacyc is already (corretly) annotated
        if (
            "metacyc.reaction" in r.annotation
            and len(data[data["Reactions"] == r.annotation["metacyc.reaction"]]) != 0
        ):
            direction = data[data["Reactions"] == r.annotation["metacyc.reaction"]][
                "Reaction-Direction"
            ].iloc[0]
            r.notes["BioCyc direction check"] = f"found {direction}"
        # complicated case: no metacyc annotation
        else:
            annotations = []

            # collect matches
            if (
                "kegg.reaction" in r.annotation
                and r.annotation["kegg.reaction"] in data["KEGG reaction"].tolist()
            ):
                annotations.append(
                    data[data["KEGG reaction"] == r.annotation["kegg.reaction"]][
                        "Reactions"
                    ].tolist()
                )
            if (
                "metanetx.reaction" in r.annotation
                and r.annotation["metanetx.reaction"] in data["METANETX"].tolist()
            ):
                annotations.append(
                    data[data["METANETX"] == r.annotation["metanetx.reaction"]][
                        "Reactions"
                    ].tolist()
                )
            if (
                "ec-code" in r.annotation
                and r.annotation["ec-code"] in data["EC-Number"].tolist()
            ):
                annotations.append(
                    data[data["EC-Number"] == r.annotation["ec-code"]][
                        "Reactions"
                    ].tolist()
                )

            # check results
            # no matches
            if len(annotations) == 0:
                r.notes["BioCyc direction check"] = "not found"

            # matches found
            else:
                # built intersection
                intersec = set(annotations[0]).intersection(*annotations)
                # case 1: exactly one match remains
                if len(intersec) == 1:
                    entry = intersec.pop()
                    direction = data[data["Reactions"] == entry][
                        "Reaction-Direction"
                    ].iloc[0]
                    r.annotation["metacyc.reaction"] = entry
                    r.notes["BioCyc direction check"] = f"found {direction}"

                # case 2: multiple matches found -> inconclusive
                else:
                    r.notes["BioCyc direction check"] = f"found, but inconclusive"

        # update direction if possible and needed
        if not pd.isnull(direction):
            if "REVERSIBLE" in direction:
                # set reaction as reversible by setting default values for upper and lower bounds
                r.lower_bound = cobra.Configuration().lower_bound
            elif "RIGHT-TO-LEFT" in direction:
                # invert the default values for the boundaries
                r.lower_bound = cobra.Configuration().lower_bound
                r.upper_bound = 0.0
            elif "LEFT-To-RIGHT" in direction:
                # In case direction was already wrong
                r.lower_bound = 0.0
                r.upper_bound = cobra.Configuration().upper_bound
            else:
                # left to right case is the standard for adding reactions
                # = nothing left to do
                continue

    return model


# Perform all clean-up steps
# --------------------------
def polish_model(
    model: libModel,
    id_db: str = "BiGG",
    mapping_tbl_file: str = None,
    gff_paths: list[str] = None,
    email: str = None,
    contains_locus_tags: bool = False,
    lab_strain: bool = False,
    kegg_organism_id: str = None,
    reaction_direction: str = None,
    outpath: str = None,
) -> libModel:
    """Completes all steps to polish a model

    .. note::

        So far only tested for models having either BiGG or VMH identifiers.

    Args:
        - model (libModel):
             Model loaded with libSBML
        - id_db (str, optional):
            Main database where identifiers in model come from.
            Defaults to 'BiGG'.
        - mapping_tbl_file (str, optional):
            Path to a file containing a mapping table with columns ``model_id | X...`` where X can be ``REFSEQ``,
            ``NCBI``, ``locus_tag`` or ``UNCLASSIFIED``.
            The table can contain all of the ``X`` columns or at least one of them.
            Defaults to None.
        - gff_paths (list[str], optional):
            Path(s) to GFF file(s). Allowed GFF formats are: RefSeq, NCBI and Prokka.
            This is only used when mapping_tbl_file == None.
            Defaults to None.
        - email (str, optional):
            E-mail for NCBI queries.
            This is only used when mapping_tbl_file == None.
            Defaults to None.
        - contains_locus_tags (bool, optional):
            Specifies if provided model has locus tags within the label tag if set to True.
            This is only used when mapping_tbl_file == None.
            Defaults to False.
        - lab_strain (bool, optional):
            Specifies if a strain from no database was provided and thus has only homolog mappings, if set to True.
            Defaults to False.
        - kegg_organism_id (str, optional):
            KEGG organism identifier if available.
            Defaults to None.
        - reaction_direction (str, optional):
            Path to a CSV file containing the BioCyc smart table with the columns
            ``Reactions (MetaCyc ID) | EC-Number | KEGG reaction | METANETX | Reaction-Direction``.
            For more details see :py:func:`~refinegems.curation.curate.check_direction`
            Defaults to None.
        - outpath (str, optional):
            Output path for mapping table from model ID to valid database IDs (if mapping_tbl_file == None)
            & incorrect annotations file(s).
            Defaults to None.

    Returns:
        libModel:
            Polished libSBML model
    """
    ### Set-up
    # Get ListOf objects
    metab_list = model.getListOfSpecies()
    reac_list = model.getListOfReactions()
    gene_list = model.getPlugin("fbc").getListOfGeneProducts()

    ### unit definition ###
    polish_model_units(model)
    set_model_default_units(model)
    set_units_of_parameters(model)
    add_compartment_structure_specs(model)
    set_initial_amount_metabs(model)

    ### improve metabolite, reaction and gene annotations ###
    extend_metab_reac_annots_via_id(metab_list, id_db)
    extend_metab_reac_annots_via_id(reac_list, id_db)
    extend_metab_reac_annots_via_notes(metab_list)
    extend_metab_reac_annots_via_notes(reac_list)
    update_annotations_from_others(model)

    ### Extend annotations for GeneProducts ###
    extend_gp_annots_via_mapping_table(
        model,
        mapping_tbl_file,
        gff_paths,
        email,
        contains_locus_tags,
        lab_strain,
        outpath
    )
    if kegg_organism_id:
        extend_gp_annots_via_KEGG(gene_list, kegg_organism_id)

    ### Check reaction direction ###
    if reaction_direction:
        check_direction(model, reaction_direction)

    ### set boundaries and constants ###
    polish_entity_conditions(metab_list)
    polish_entity_conditions(reac_list)

    ### MIRIAM compliance of CVTerms ###
    print(
        "Remove duplicates & transform all CURIEs to the new identifiers.org pattern (: between db and ID):"
    )
    model = polish_annotations(model, True, outpath)
    print("Changing all qualifiers to be MIRIAM compliant:")
    model = change_all_qualifiers(model, lab_strain)

    return model



"""Collection of functions for setting up files and work environments."""

__author__ = "Carolin Brune"

################################################################################
# requirements
################################################################################

import requests
import subprocess
import warnings

from importlib.resources import files
from pathlib import Path
from tqdm import tqdm
from typing import Literal

################################################################################
# variables
################################################################################

PATH_MEDIA_CONFIG = files("refinegems.data.config").joinpath(
    "media_config.yml"
)  #: :meta hide-value:

################################################################################
# functions
################################################################################

# --------------------------
# download databases / files
# --------------------------


def download_url(
    download_type: Literal["SwissProt gapfill"],
    directory: str = None,
    k: int = 10,
    t: int = 1,
):
    """Download files necessary for certain functionalities of the toolbox from
    the internet.

    Currently available:

    - 'SwissProt gapfill': download files needed for the :py:class:`~refinegems.classes.gapfill.GeneGapFiller`


    Args:
        - dowload_type (Literal['SwissProt gapfill']):
            Type of files to download.
        - directory (str, optional):
            Path to a directory to save the downloaded files to.
            Defaults to None (So the current working directory is used).
        - k (int, optional):
            Chunksize in kB.
            Defaults to 10.
        - t (int, optional):
            Number of threads to use for some additional setups, e.g.
            DIAMOND database creation.
            Defaults to 1.

    Raises:
        - ValueError: Unknown database or file
    """

    # Ensure that directory is not None
    if directory is None:
        directory = Path.cwd()

    # match URLS to type of database, that the user wants to download
    match download_type:
        case "SwissProt gapfill":
            swissprot_api = "https://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.fasta.gz"
            swissprot_mapping_api = "https://rest.uniprot.org/uniprotkb/stream?fields=accession%2Cxref_brenda%2Cec&format=tsv&query=%28*%29+AND+%28reviewed%3Atrue%29"
            urls = {
                "SwissProt.fasta": swissprot_api,
                "SwissProt_mapping.tsv": swissprot_mapping_api,
            }
            #   1.: TSV with UniprotID, BRENDA and EC -7.7MB (26.07.2024)
            #   2.: FASTA with sequences ~280MB (26.07.2024)
        case _:
            mes = f"Unknown database or file: {download_type}"
            raise ValueError(mes)

    # download each file
    for name, url in urls.items():
        r = requests.get(url, stream=True)
        filename = Path(directory, name)

        # Check if Content-Length is available
        total_length = r.headers.get("Content-Length")  # Make the progress bar
        if total_length is None:
            # Content-Length is missing, so we download without a progress bar
            with open(filename, "wb") as f:
                for chunk in r.iter_content(chunk_size=k * 1024):
                    if chunk:
                        f.write(chunk)
        else:
            total_length = int(total_length)
            with open(filename, "wb") as f:
                pbar = tqdm(
                    desc=f"Downloading {name}",
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                    total=total_length,
                )
                pbar.clear()
                for chunk in r.iter_content(chunk_size=k * 1024):
                    if chunk:  # Filter out keep-alive new chunks
                        pbar.update(len(chunk))  # Update progress bar
                        f.write(chunk)
                pbar.close()

    # additional setups
    match download_type:
        # SwissProt gapfill
        case "SwissProt gapfill":
            # Create db folder for DIAMONd if non-existent
            try:
                Path(directory, "db").mkdir(parents=True, exist_ok=False)
                print("Creating new directory " + str(Path(directory, "db")))
            except FileExistsError:
                warnings.warn(
                    "Given directory already exists. High possibility of files being overwritten."
                )
            Path(directory, "db").mkdir(parents=True, exist_ok=True)
            # create DIAMOND database
            print(f"create DIAMOND database for {download_type} using:")
            print(
                f'diamond makedb --in {Path(directory, "SwissProt.fasta")} --db {str(Path(directory,"db","SwissProt.dmnd"))} --threads {int(t)}'
            )
            subprocess.run(
                [
                    "diamond",
                    "makedb",
                    "--in",
                    str(Path(directory, "SwissProt.fasta")),
                    "--db",
                    str(Path(directory, "db", "SwissProt.dmnd")),
                    "--threads",
                    str(t),
                ]
            )
        # Type for which no extra setup is needed
        case _:
            pass


# ---------------------
# handling config files
# ---------------------


def download_config(filename: str = "./my_config.yaml", type=Literal["media"]):
    """Load a configuration file from the package and save a copy of it for the user to edit.

    Args:
        - filename (str, optional):
            Filename to write the config to/save it under as.
            Defaults to './my_config.yaml'.
        - type (Literal['media'], optional):
            Type of configuration file to load.
            Can be 'media' for the media config file.
            Defaults to Literal['media'].
    """

    def copy_config_yaml(infile: str, outfile: str):
        """Helper function for :py:func:`download_config`.
        Performs the actual download of a yaml file into a copy for the user to edit.

        Args:
            - infile (str):
                Path to the file to copy.
            - outfile (str):
                Path to write the copy to.
        """
        with open(infile, "r") as cfg_file, open(outfile, "w") as cfg_out:
            for line in cfg_file:
                cfg_out.write(line)

    # copy an examplary version of the config file for the user to edit it
    match type:

        # copy media config
        case "media":
            copy_config_yaml(PATH_MEDIA_CONFIG, filename)

        # type not found
        case _:
            raise ValueError(f"Unknown type of config file detected: {type}")



"""Identify, report and solve energy generating cycles (EGCs)."""

__author__ = "Carolin Brune and Tobias Fehrenbach"

################################################################################
# requirements
################################################################################

from ..analysis.growth import set_bounds_to_default
from .medium import Medium
from ..utility.util import test_biomass_presence, MIN_GROWTH_THRESHOLD

import cobra
import pandas as pd
import warnings

from functools import partial
from multiprocess import Pool
from tqdm import tqdm
from typing import Literal, Union

################################################################################
# variables
################################################################################

DISSIPATION_RXNS = {
    "ATP": {
        "ATP [Adenosine triphosphate]": -1,
        "Water [H2O]": -1,
        "ADP [Adenosine diphosphate]": 1,
        "Hydrogen [H(+)]": 1,
        "Phosphate [PO4(3-)]": 1,
    },
    "CTP": {
        "CTP [Cytidine triphosphate]": -1,
        "Water [H2O]": -1,
        "CDP [Cytidine diphosphate]": 1,
        "Hydrogen [H(+)]": 1,
        "Phosphate [PO4(3-)]": 1,
    },
    "GTP": {
        "GTP [Guanosine triphosphate]": -1,
        "Water [H2O]": -1,
        "GDP [Guanosine diphosphate]": 1,
        "Hydrogen [H(+)]": 1,
        "Phosphate [PO4(3-)]": 1,
    },
    "UTP": {
        "UTP [Uridine triphosphate]": -1,
        "Water [H2O]": -1,
        "UDP [Uridine diphosphate]": 1,
        "Hydrogen [H(+)]": 1,
        "Phosphate [PO4(3-)]": 1,
    },
    "ITP": {
        "ITP [Inosine triphosphate]": -1,
        "Water [H2O]": -1,
        "IDP [Inosine diphosphate]": 1,
        "Hydrogen [H(+)]": 1,
        "Phosphate [PO4(3-)]": 1,
    },
    "NADH": {
        "NADH [reduced Nicotinamide adenine dinucleotide]": -1,
        "Hydrogen [H(+)]": 1,
        "NAD [oxidized Nicotinamide adenine dinucleotide]": 1,
    },
    "NADPH": {
        "NADPH [reduced Nicotinamide adenine dinucleotide phosphate]": -1,
        "Hydrogen [H(+)]": 1,
        "NADP [oxidized Nicotinamide adenine dinucleotide phosphate]": 1,
    },
    "FADH2": {
        "FADH2 [reduced Flavin adenine dinucleotide]": -1,
        "Hydrogen [H(+)]": 2,
        "FAD [oxidized Flavin adenine dinucleotide]": 1,
    },
    "FMNH2": {
        "FMNH2 [reduced Flavin mononucleotide]": -1,
        "Hydrogen [H(+)]": 2,
        "FMN [oxidized Flavin mononucleotide]": 1,
    },
    "Q8H2": {"Ubiquinone-8": -1, "Hydrogen [H(+)]": 2, "Ubiquinol-8": 1},
    "MQL8": {"Menaquinone-8": -1, "Hydrogen [H(+)]": 2, "Menaquinol-8": 1},
    "DMMQL8": {
        "2-Demethylmenaquinone-8": -1,
        "Hydrogen [H(+)]": 2,
        "2-Demethylmenaquinol-8": 1,
    },
    "ACCOA": {
        "Acetyl-CoA": -1,
        "Water [H2O]": -1,
        "Hydrogen [H(+)]": 1,
        "Acetate [Acetic acid]": 1,
        "Coenzyme A": 1,
    },
    "GLU": {
        "D-Glucose": -1,
        "Water [H2O]": -1,
        "2-Oxoglutarate [Oxoglutaric acid]": 1,
        "Ammonia": 1,
        "Hydrogen [H(+)]": 2,
    },
    "PROTON": {"Hydrogen [H(+)]": 1, "Hydrogen [H(+)] transported": -1},
}  #: :meta:

EGC_SCORING_MATRIX = {"MR": 1, "RB": 3, "RF": 3, "RM": 6}  #: :meta:

################################################################################
# classes
################################################################################


class EGCSolver:
    """Parent class for the EGC solvers with generally useful functions and
    attributes. Can only be used to find, not solve EGCs directly.

    Attributes:
        - theshold: Float describing the cutoff, under which the model
            will no longer considered to be growing.
            Defaults to the MIN_GROWTH_THRESHOLD set in the growth module.
        - limit: Sets the maximal number of cores to be used.
            Defaults to 2.
        - chunksize: Chunksize to use for multiprocessing.
            Defaults to 1.
    """

    def __init__(
        self,
        threshold: float = MIN_GROWTH_THRESHOLD,
        limit: int = 2,
        chunksize: int = 1,
    ) -> None:

        # 'biological' attributes
        self.threshold = threshold

        # purely computational attributes
        self.limit = limit
        self.chunksize = chunksize

    # general functions
    # -----------------
    def check_metab_integration(
        self,
        metabolites: dict[str:int],
        model: cobra.Model,
        metab_info: Medium,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> Union[None, dict]:
        """Check if the metabolites of a reactions are in the model.
        If yes, return the dictionary of metabolites (their IDs in the model) to the factors.
        If no, return None

        Args:
            - metabolites (dict[str: int]):
                Metabolites mapped to factors.
            - model (cobra.Model):
                The model loaded with COBRApy.
            - metab_info (Medium):
                Information about the metabolites from the database,
                in for of a Medium object.
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the
                compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            (1) Case: metabolites not found
                    None:
                        nothing to return

            (2) Case: found
                    dict:
                        The mapping of IDs of the metabolites to the factors.
        """

        found_ids = {}

        c1_metab = [_.id for _ in model.metabolites if _.compartment == compartment[0]]
        c2_metab = [_.id for _ in model.metabolites if _.compartment == compartment[1]]

        for meta in list(metabolites.keys()):
            # get metabolite database annotations
            pos_ids = metab_info.substance_table[
                metab_info.substance_table["name"] == meta
            ][["db_type", "db_id"]]
            # check namespace availability
            if not any(namespace in _ for _ in list(pos_ids["db_type"])):
                return None
            # check metabolite availability in model
            id = pos_ids[pos_ids["db_type"].str.contains(namespace)]["db_id"]
            for i in id:
                match namespace:
                    # BiGG ID needs the compartment with a '_' as suffix
                    case "BiGG":
                        if len(metabolites) == 2:  # special case proton
                            i_p = i + "_" + compartment[1]
                            i_c = i + "_" + compartment[0]
                            if i_p not in c2_metab or i_c not in c1_metab:
                                return None
                            else:
                                found_ids[i_p] = metabolites[
                                    "Hydrogen [H(+)] transported"
                                ]
                                found_ids[i_c] = metabolites["Hydrogen [H(+)]"]
                                return found_ids
                        else:
                            i += "_" + compartment[0]
                    # No namespacce given
                    case _:
                        mes = f"Unknown namespace or no namespace given: {namespace}"
                        warnings.warn(mes)
                        return None

                if i not in c1_metab:
                    return None
                else:
                    found_ids[i] = metabolites[meta]
                    break

        return found_ids

    def add_DISSIPATIONRXNS(
        self,
        model: cobra.Model,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> cobra.Model:
        """Add the dissipation reactions a model.

        Args:
            - model (cobra.Model):
                A model loaded with COBRApy.
            - namespace (Literal['BiGG'], optional): Namespace of the model.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            cobra.Model:
                The edited model.
        """

        # retrieve information about dissipation reaction metabolites
        metab_info = Medium("dissipation reaction metabolites")
        metab_info.add_subset("DiReM")

        # add dissipations reactions to model
        for name, metabolites in DISSIPATION_RXNS.items():
            if "DISSI_" + name not in model.reactions:
                ids_in_namespace = self.check_metab_integration(
                    metabolites,
                    model,
                    metab_info,
                    namespace=namespace,
                    compartment=compartment,
                )
                if ids_in_namespace:
                    rea = cobra.Reaction(name=name, id="DISSI_" + name)
                    model.add_reactions([rea])
                    rea.add_metabolites(ids_in_namespace)

        return model

    def limit_bounds(self, model: cobra.Model):
        """Limits upper and lower bounds of

        - exchange reactions to (0, 0)
        - reversible reactions to (-1, 1)
        - irreversible reactions to (0, 1)

        Excludes dissipation reactions.

        Args:
            - model (cobra.Model):
                COBRApy model
        """
        external_comp = cobra.medium.find_external_compartment(model)
        # set fluxes for each reaction within model
        for rea in model.reactions:
            # except dissipation reactions
            if "DISSI_" in rea.id:
                continue
            # turn off exchange reactions
            elif cobra.medium.is_boundary_type(rea, "exchange", external_comp):
                rea.bounds = (0.0, 0.0)
            # limit reversible reactions to [-1, 1] -> flux 0.1
            elif rea.reversibility:
                rea.bounds = (-1.0, 1.0)
            # limit irreversible reactions to [0, 1] -> flux 0.1
            else:
                rea.bounds = (0.0, 1.0)

    # find EGCs
    # ---------

    def find_egcs(
        self,
        model: cobra.Model,
        with_reacs: bool = False,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> Union[list, tuple]:
        """Find the EGCs in a model - if exsistend.

        Args:
            - model (cobra.Model):
                The model loaded with COBRApy.
            - with_reacs (bool, optional):
                Option to either only return the names
                of the found EGC or additionally also the reactions, which
                show fluxes during testing.
                Defaults to False.
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the
                compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            (1) Case: ``with_reacs = False``
                    list:
                        List of found EGC names.

            (2) Case: ``with_reacs = True``
                    tuple:
                        tuple of (1) dictionary & (2) list:

                        (1) dict: dictionary of the EGCs
                        (2) list: their reactions that showed fluxes and the objective values of the test.
        """
        # add 15 energy dissipation reactions
        with model as mod_model:

            # ensure no model modifications
            mod_model = self.add_DISSIPATIONRXNS(mod_model, namespace, compartment)

            # set fluxes for each reaction within model
            self.limit_bounds(mod_model)

            # set each dissipation reaction as objective & optimize
            # options: just return the names of the found cycles
            if not with_reacs:
                egcs = []
                for name in DISSIPATION_RXNS.keys():
                    rea_id = "DISSI_" + name
                    if rea_id in mod_model.reactions:
                        mod_model.objective = "DISSI_" + name
                        solution = mod_model.optimize()

                        if (
                            solution.objective_value > MIN_GROWTH_THRESHOLD
                        ):  # optimization > 0 --> EGC detected
                            egcs.append(name)

                return egcs
            # option 2: report cycles + problematic reactions with fluxes
            else:
                egc_reactions = {}
                obj_vals = {}
                for name in DISSIPATION_RXNS.keys():
                    rea_id = "DISSI_" + name
                    if rea_id in mod_model.reactions:
                        mod_model.objective = "DISSI_" + name
                        solution = mod_model.optimize()
                        fluxes = solution.fluxes
                        objval = solution.objective_value

                        if (
                            objval > MIN_GROWTH_THRESHOLD
                        ):  # optimization > 0 --> EGC detected
                            obj_vals[name] = objval
                            egc_reactions[name] = {}

                            for rea, flux in fluxes.items():
                                # cutoff for flux
                                if (
                                    flux > MIN_GROWTH_THRESHOLD
                                    or flux < -1.0 * MIN_GROWTH_THRESHOLD
                                ) and not rea.startswith("DISSI"):
                                    egc_reactions[name][rea] = flux

                return (obj_vals, egc_reactions)

    # Helper functions for testing during solving process
    # ---------------------------------------------------

    def egcs_removed(
        self,
        model: cobra.Model,
        starting_egcs: dict,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> list:
        """Compare a list of previously found EGCs to the current
        EGCs in the model.

        Args:
            - model (cobra.Model):
                The model loaded with COBRApy after
                a try of solving the EGCs.
            - starting_egcs (dict):
                List of EGCs before trying to solve them.
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            list:
                List of newly removed EGCs.
        """

        current_egcs = self.find_egcs(
            model, namespace=namespace, compartment=compartment
        )

        removed = [_ for _ in starting_egcs.keys() if _ not in current_egcs]
        added = [_ for _ in current_egcs if _ not in starting_egcs.keys()]

        if len(added) > 0:

            return []
        else:
            return removed


class GreedyEGCSolver(EGCSolver):
    """EGC solver that finds a good solution (greedy) based on modifications
    to single reactions.

    Workflow:

    - identify existing EGCs
    - test, if EGCs can be solved using single modifications of reactions

        - possible modifications:

            - deletion (RM)
            - set reversible (MR)
            - remove backward (forward only) (RB)
            - remove forward (backward only) (RF)

    - find a good - not optimal - combination of reactions, that solve
      the maximum number of EGCs that can be solved this way
    - apply solution to the model
    - report remaining EGCs, score and reactions used for solution

    Attributes:
        - all attributes of the base class :py:class:`refinegems.classes.egcs.EGCSolver`
        - scoring_matrix:
            Dictionary of the changes (RM, MR, RF, RB) against
            Integers describing the penalty scores.
    """

    def __init__(self, scoring_matrix: dict = EGC_SCORING_MATRIX, **kwargs) -> None:
        super().__init__(**kwargs)
        self.scoring_matrix = scoring_matrix

    # find modifications to solve single EGCs
    # ---------------------------------------

    def check_egc_growth(
        self,
        reac: cobra.Reaction,
        model: cobra.Model,
        bounds: tuple,
        starting_egcs: dict,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> Union[list, None]:
        """Check EGC removal and growth of a model when chaning the bounds
        of a single reaction.

        Args:
            - reac (cobra.Reaction):
                The reaction to change
            - model (cobra.Model):
                The model (COBRApy) to manipulate.
            - bounds (tuple):
                The new reactions bounds.
            - starting_egcs (dict):
                Dict of the original EGCs found in the model.
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional): List of length 2 with the names of the
                compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            (1) Case if EGCs removed
                    list:
                        List of EGCs that can be removed with the change

            (2) Case no removal possible
                    None:
                        no return
        """

        # set new reaction bounds
        reac.bounds = bounds
        # check egcs removal and growth
        egc_test = self.egcs_removed(model, starting_egcs, namespace, compartment)
        res = model.optimize()
        if (
            len(egc_test) > 0
            and res.status == "optimal"
            and res.objective_value > self.threshold
        ):
            return egc_test
        else:
            None

    def test_modifications(
        self,
        reaction: cobra.Reaction,
        model: cobra.Model,
        present_egc: dict,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> dict:
        """Tries four cases for a Reaction

        1. if reaction is not reversible -> make reaction reversible (MR)
        2. limit backward reaction (RB)
        3. limit forward reaction (RF)
        4. "delete" reaction by setting fluxes to 0 (RM)

        -> for each case the EGCs which are present in the model are checked if they are removed

        -> if EGCs are removed we check if the model still grows on optimal medium

        => When both limitations are True reaction is saved to corresponding dictionary


        Args:
            - reaction (cobra.Reaction):
                Reaction from a cobra.Model
            - model (cobra.Model):
                The corresponding GEM loaded with cobrapy
            - present_egc (dict):
                Dictionary of present EGCs {"egc": {}} -> EGCs are keys
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the
                compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            dict:
                {"egc": {"MR":[potential_solutions],
                "RB":[potential_solutions],
                "RF":[potential_solutions],
                "RM":[potential_solutions]}}
        """
        results = {
            "ATP": {"MR": [], "RB": [], "RF": [], "RM": []},
            "CTP": {"MR": [], "RB": [], "RF": [], "RM": []},
            "GTP": {"MR": [], "RB": [], "RF": [], "RM": []},
            "UTP": {"MR": [], "RB": [], "RF": [], "RM": []},
            "ITP": {"MR": [], "RB": [], "RF": [], "RM": []},
            "NADH": {"MR": [], "RB": [], "RF": [], "RM": []},
            "NADPH": {"MR": [], "RB": [], "RF": [], "RM": []},
            "FADH2": {"MR": [], "RB": [], "RF": [], "RM": []},
            "FMNH2": {"MR": [], "RB": [], "RF": [], "RM": []},
            "Q8H2": {"MR": [], "RB": [], "RF": [], "RM": []},
            "MQL8": {"MR": [], "RB": [], "RF": [], "RM": []},
            "DMMQL8": {"MR": [], "RB": [], "RF": [], "RM": []},
            "ACCOA": {"MR": [], "RB": [], "RF": [], "RM": []},
            "GLU": {"MR": [], "RB": [], "RF": [], "RM": []},
            "PROTON": {"MR": [], "RB": [], "RF": [], "RM": []},
        }

        # skip exchange reactions
        if reaction.boundary:
            return None

        # test modifications of the reaction
        with model as m:
            original_bounds = reaction.bounds
            # case 1: make reaction reversible
            if not reaction.reversibility:  # skip if reaction is already reversible
                res = self.check_egc_growth(
                    reaction,
                    m,
                    (-1000.0, 1000.0),  # irreversible -> reversible
                    present_egc,
                    namespace,
                    compartment,
                )
                if res:
                    for egc in res:
                        results[egc]["MR"].append(reaction.id)  # MR = make reversible

            # case 2: limit backward reaction
            if original_bounds != (0.0, 1000.0):
                res = self.check_egc_growth(
                    reaction,
                    m,
                    (0.0, 1000.0),  # Remove Backward (RB) reaction
                    present_egc,
                    namespace,
                    compartment,
                )
                if res:
                    for egc in res:
                        results[egc]["RB"].append(reaction.id)

            # case 3: limit forward reaction
            if original_bounds != (-1000.0, 0):
                res = self.check_egc_growth(
                    reaction,
                    m,
                    (-1000.0, 0),  # Remove Forward (RF) reaction
                    present_egc,
                    namespace,
                    compartment,
                )
                if res:
                    for egc in res:
                        results[egc]["RF"].append(reaction.id)

            # case 4: "delete" reaction
            if original_bounds != (0.0, 0.0):
                res = self.check_egc_growth(
                    reaction,
                    m,
                    (0.0, 0.0),  # Remove (RM) reaction
                    present_egc,
                    namespace,
                    compartment,
                )
                if res:
                    for egc in res:
                        results[egc]["RM"].append(reaction.id)

        return results

    def find_mods_resolve_egcs_greedy(
        self,
        model: cobra.Model,
        present_egcs: dict,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> dict:
        """Find the (single) modifications to reactions in a cobra.Model and returns these in a dictionary.
        Splits the modification check in multiple processes.

        Args:
            - model (cobra.Model):
                The model loaded with COBRApy.
            - present_egcs (dict):
                Dict of the original EGCs found in the model.
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the
                compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            dict:
                Dictionary of potential modifications to resolve EGCs
                {"egc": {"MR":[potential_solutions],
                "RB":[potential_solutions],
                "RF":[potential_solutions],
                "RM":[potential_solutions]}}
        """

        output_list = []

        if len(present_egcs) == 0:
            print("No EGCs present, nothing to solve.")

        else:
            print("_____________________________________")
            print(
                f"Try to resolve the following EGCs: {[egc for egc in present_egcs]} \nThis might take a while..."
            )

            # try to find BOF
            pos_bofs = test_biomass_presence(model)
            if pos_bofs:
                bof = pos_bofs[0]
            else:
                mes = "No growth or biomass objectuve function in model. Cannot solve EGCs."
                raise KeyError(mes)

            with model as m:
                set_bounds_to_default(m)
                m.objective = bof

                # might limit processes with Pool(process=limit) -> otherwise it consumes all cores
                # limit to half of machine cores -> at least for me no speed increment with more cores
                try:
                    pool = Pool(processes=self.limit)

                    # partial -> creates function with fixed variables to call with each iteration
                    # needed since pool.imap cannot do that
                    part_test_mods = partial(
                        self.test_modifications,
                        model=m,
                        present_egc=present_egcs,
                        namespace=namespace,
                        compartment=compartment,
                    )
                    # increment in chunksize will reduce computation time -> but progressbar update also...
                    for res in list(
                        tqdm(
                            pool.imap_unordered(
                                func=part_test_mods,
                                iterable=m.reactions,
                                chunksize=self.chunksize,
                            ),
                            total=len(m.reactions),
                            desc="Resolve EGCs",
                        )
                    ):
                        if res:
                            output_list.append(res)
                finally:
                    pool.close()
                    pool.join()

            # merge output_list to the final results
            results = {}
            for output_dict in output_list:
                for egc, modifications in output_dict.items():
                    for mod, reactions in modifications.items():
                        if not egc in results.keys():
                            results[egc] = {}
                        if mod in results[egc].keys():
                            results[egc][mod] = list(set(results[egc][mod] + reactions))
                        else:
                            results[egc][mod] = list(set(reactions))

            return results

    # find a solution to solve all EGCs that
    # have a solution based on previous step
    # --------------------------------------
    def find_solution_greedy(self, results: dict, egc_reactions: dict) -> tuple:
        """Based on the originally found EGCs and the output of
        :py:func:`find_mods_resolve_egcs_greedy`, find a solution that
        solves all EGCs that can be solved with the results.

        Args:
            - results (dict):
                Output of :py:func:`find_mods_resolve_egcs_greedy`.
            - egc_reactions (dict):
                Output of :py:meth:`~refinegems.classes.egcs.EGCSolver.find_egcs` with 'with_reac=True'.
                Should be the EGCs before calculating any solutions and applying them.
            - scoring_matrix (dict, optional):
                Dictionary of the modifications types
                (RM, MR, RF, RB) and their penality score.
                Defaults to the in-build scoring matrix.

        Returns:
            tuple:
                Tuple of (1) dict & (2) int:

                (1) dictionary of reaction IDs and their mode of change
                (2) score of the solution
        """

        solved_egcs = set()
        score = 0
        reacs_for_solu = {}

        # make table and drop EGCs that are not in the model
        solution_table = pd.DataFrame(results)[list(egc_reactions.keys())]
        for col in solution_table.columns:
            solution_table[col] = solution_table[col].apply(
                lambda x: ", ".join(map(str, x))
            )
        # find not solvable (with the current code) cyles
        not_solvable = []
        for col in egc_reactions.keys():
            if all(len(_) == 0 for _ in solution_table[col]):
                not_solvable.append(col)
        # report the problems for manual curation
        if len(not_solvable) > 0:
            print(
                f"The following EGCs cannot be fixed with the current code: {not_solvable}"
            )
            solution_table.drop(not_solvable, axis=1, inplace=True)
        # fix the EGCs that can be fixed with single changes
        solution_table = solution_table.T
        for egc in solution_table.index.to_list():
            # check if sgc has already been taken care of
            if egc in solved_egcs:
                continue

            # best solution for this cycle
            for mode in self.scoring_matrix.keys():
                if solution_table.loc[egc, mode]:
                    reac = solution_table.loc[egc, mode].split(", ")[0]
                    newly_solved = [
                        _
                        for _ in solution_table.index[
                            solution_table[mode].str.contains(reac)
                        ].to_list()
                        if _ not in solved_egcs
                    ]
                    if len(newly_solved) > 0:
                        solved_egcs = set([*solved_egcs, *newly_solved])
                        score += self.scoring_matrix[mode]
                        reacs_for_solu[reac] = mode

        return (reacs_for_solu, score)

    # apply the found solution the model
    # ----------------------------------
    def apply_modifications(self, model: cobra.Model, solution: dict):
        """Apply the modifications to reactions in solution to the model.

        4 modifications are possible:

        - "RM" -> removes the reaction
        - "RB" -> removes the backwards reaction
        - "RF" -> removes the forward reaction
        - "MR" -> makes reaction reversible

        Args:
            - model (cobra.Model):
                Input model
            - solution (dict):
                Best solution from calculation in `py:func:find_solution_greedy`.
        """
        for reac, mode in solution.items():
            reaction = model.reactions.get_by_id(reac)
            if type(reaction) is cobra.Reaction:
                match mode:
                    case "RM":
                        reaction.delete()
                    case "RB":
                        reaction.bounds = (0.0, 1000.0)
                    case "RF":
                        model = reaction.bounds = (1000.0, 0.0)
                    case "MR":
                        model = reaction.bounds = (1000.0, 1000.0)
                    case _:
                        mes = (
                            f"{mode} is no viable modification... Something went wrong."
                        )
                        warnings.warn(mes)

    # run the complete solving process
    # --------------------------------
    def solve_egcs(
        self,
        model: cobra.Model,
        namespace: Literal["BiGG"] = "BiGG",
        compartment: list = ["c", "e"],
    ) -> Union[dict, None]:
        """Run the complete greedy EGC solving process.

        Note: The input model gets changed, if EGCs can be solved.

        Args:
            - model (cobra.Model):
                The model loaded with COBRApy.
            - namespace (Literal['BiGG'], optional):
                String for the namespace used in the model.
                Current options include 'BiGG'.
                Defaults to 'BiGG'.
            - compartment (list, optional):
                List of length 2 with the names of the
                compartments for the dissipations reactions.
                Defaults to ['c','e'].

        Returns:
            dict:
                Dictionary with the following entries.

                - 'solution': List of reactions for the solution.
                - 'score': Score of the solution.
                - 'remaining egcs': List of EGCs that could not be solved.

        """
        egc_reactions, obj_vals = self.find_egcs(
            model, with_reacs=True, namespace=namespace, compartment=compartment
        )
        results = self.find_mods_resolve_egcs_greedy(
            model=model,
            present_egcs=egc_reactions,
            namespace=namespace,
            compartment=compartment,
        )
        if results:
            solution, score = self.find_solution_greedy(results, egc_reactions)
            self.apply_modifications(model, solution)
            still_egc = self.find_egcs(
                model, namespace=namespace, compartment=compartment
            )

            return {"solution": solution, "score": score, "remaining egcs": still_egc}



#!/usr/bin/env python
"""This module contains decorator functions."""

__author__ = "Carolin Brune"

################################################################################
# requirements
################################################################################

import logging
from functools import wraps, partial

################################################################################
# variables
################################################################################

################################################################################
# functions
################################################################################


def template(func):
    """A decorator for template functions.

    Template functions are not executable and are mainly for giving developers
    a sense of how functions for the functionality this template is for should look like.
    """

    @wraps(func)
    def wrapper():
        raise RuntimeError(
            "This function is a template for developers.\nIt cannot be executed."
        )

    return wrapper


def implement(func):
    """A decorator for functions that need to be implemented.

    Used to give a hint to other developers, that this function is not yet implemented,
    but should be to make the program executable.
    """

    @wraps(func)
    def wrapper():
        raise NotImplementedError(
            "The current function is just a placeholder and will be implement in the fucture."
        )

    doc_extension = """\n.. note:: \n\t*Will be coming in a future release.*"""
    if wrapper.__doc__:
        wrapper.__doc__ += doc_extension
    else:
        wrapper.__doc__ = doc_extension
    return wrapper


def debug(func):
    """A decorator for debug functions.

    Debug functions are for giving developers detailed output for debugging the code.
    """

    @wraps(func)
    def wrapper():
        raise RuntimeError("This function is for debugging for the developers.")

    return wrapper


def deprecate(func=None, note: str = None):
    """A decorator to tell the user, that the function will soon be deprecated.

    Used to give hints to users, that their code will not work as expected, if they update
    to a newer version.
    """

    if func is None:
        return partial(deprecate, note=note)

    @wraps(func)
    def wrapper(*args, **kwargs):
        mes = f"This function will be deprecated in the next major update: {func.__name__}"
        if note:
            mes = mes + "\n" + f"Additional notes: {note}"
        logging.warning(mes)
        return func(*args, **kwargs)

    return wrapper



#!/usr/bin/env python
"""Helper module to work with annotations (CVTerms)

Stores dictionaries which hold information the identifiers.org syntax, has functions to add CVTerms to different entities and parse CVTerms.
"""

__author__ = "Famke Baeuerle and Gwendolyn O. Döbel"

################################################################################
# requirements
################################################################################

import cobra
import logging

from bioregistry import get_identifiers_org_iri, parse_iri
from libsbml import (
    BIOLOGICAL_QUALIFIER,
    BQB_IS,
    BQB_OCCURS_IN,
    BQB_IS_HOMOLOG_TO,
    MODEL_QUALIFIER,
    BQM_IS_DESCRIBED_BY,
)
from libsbml import Unit, CVTerm, Species, Reaction, GeneProduct, Group, SBase
from typing import Union

from ..developement.decorators import debug

################################################################################
# variables
################################################################################

DB2PREFIX_METABS = {
    "BIGG": "bigg.metabolite",
    "BIOCYC": "biocyc",
    "BRENDA": "brenda",
    "CHEBI": "chebi",
    "HMDB": "hmdb",
    "HUMAN METABOLOME DATABASE": "hmdb",
    "INCHI": "inchi",
    "INCHI-KEY": "inchikey",
    "KEGG": "kegg.compound",
    "KEGG COMPOUND": "kegg.compound",
    "METACYC": "metacyc.compound",
    "MNXREF": "metanetx.chemical",
    "METANETX": "metanetx.chemical",
    "PUBCHEM": "pubchem.compound",
    "REACTOME": "reactome",
    "SEED": "seed.compound",
    #"UPA": "unipathway.compound", #closed due to financial issues
    #"UNIPATHWAY COMPOUND": "unipathway.compound"
    "VMH": "vmhmetabolite",
}  #: :meta hide-value:

PREFIX2DB_METABS = {v: k for (k, v) in DB2PREFIX_METABS.items()}  #: :meta hide-value:

DB2PREFIX_REACS = {
    "BIGG": "bigg.reaction",
    "BIOCYC": "biocyc",
    "BRENDA": "brenda",
    "EC": "ec",
    "HMDB": "hmdb",
    "HUMAN METABOLOME DATABASE": "hmdb",
    "KEGG": "kegg.reaction",
    "METACYC": "metacyc.reaction",
    "MNXREF": "metanetx.reaction",
    "METANETX": "metanetx.reaction",
    "REACTOME": "reactome",
    "RHEA": "rhea",
    "SEED": "seed.reaction",
    #"UPA": "unipathway.reaction:", #closed due to financial issues
    #"UNIPATHWAY REACTION": "unipathway.reaction:"
    "VMH": "vmhreaction",
}  #: :meta hide-value:

PREFIX2DB_REACS = {v: k for (k, v) in DB2PREFIX_REACS.items()}  #: :meta hide-value:

DB2PREFIX_GENES = {
    "KEGG": "kegg.genes",
    "NCBI": "ncbiprotein",
    "NCBIGENE": "ncbigene",
    "REFSEQ": "refseq",
    "UNIPROT": "uniprot",
}  #: :meta hide-value:

PREFIX2DB_GENES = {v: k for (k, v) in DB2PREFIX_GENES.items()}  #: :meta hide-value:

DB2PREFIX_PATHWAYS = {"KEGG": "kegg.pathway"}  #: :meta hide-value:

PREFIX2DB_PATHWAYS = {
    v: k for (k, v) in DB2PREFIX_PATHWAYS.items()
}  #: :meta hide-value:

MIRIAM = "https://identifiers.org/"  #: :meta hide-value:
OLD_MIRIAM = "http://identifiers.org/"  #: :meta hide-value:

################################################################################
# functions
################################################################################

# cobra
# -----


def _add_annotations_from_dict_cobra(
    references: dict, entity: Union[cobra.Reaction, cobra.Metabolite, cobra.Model]
) -> None:
    """Given a dictionary and a cobra object, add the former as annotations to the latter.
    The keys of the dictionary are used as the annotation labels, the values as the values.
    If the keys are already in the entity, the values will be combined (union).

    Args:
        - references (dict):
            The dictionary with the references to add the entity.
        - entity (cobra.Reaction | cobra.Metabolite | cobra.Model):
            The entity to add annotations to.
    """
    # add additional references from the parameter
    for db, idlist in references.items():
        if not isinstance(idlist, list):
            idlist = [idlist]
        if db in entity.annotation.keys():
            if isinstance(entity.annotation[db], list):
                entity.annotation[db] = list(set(entity.annotation[db] + idlist))
            else:
                entity.annotation[db] = list(set(list(entity.annotation[db]) + idlist))
        else:
            entity.annotation[db] = idlist


# libsbml
# -------


def add_cv_term_units(unit_id: str, unit: Unit, relation: int):
    """Adds CVTerm to a unit

    Args:
        - unit_id (str):
            ID to add as URI to annotation
        - unit (Unit):
            Unit to add CVTerm to
        - relation (int):
            Provides model qualifier to be added
    """
    db_id = "pubmed" if relation == BQM_IS_DESCRIBED_BY else "uo"
    resource = get_identifiers_org_iri(db_id, unit_id)
    if resource:
        cv = CVTerm()
        cv.setQualifierType(MODEL_QUALIFIER)
        cv.setModelQualifierType(relation)
        cv.addResource(resource)
        unit.addCVTerm(cv)
    else:
        logging.warning(
            f"No valid IRI could be formed for {unit} with relation {relation} and {db_id}:{unit_id}."
        )


def add_cv_term_metabolites(entry: str, db_id: str, metab: Species):
    """Adds CVTerm to a metabolite

    Args:
        - entry (str):
            Id to add as annotation
        - db_id (str):
            Database to which entry belongs. Must be in DB2PREFIX_METABS.keys().
        - metab (Species):
            Metabolite to add CVTerm to
    """
    resource = get_identifiers_org_iri(DB2PREFIX_METABS.get(db_id), entry)
    if resource:
        if db_id == "HMDB" or db_id == "HUMAN METABOLOME DATABASE":
            if entry[:4] == "HMDB":
                entry = entry[4:]
        cv = CVTerm()
        cv.setQualifierType(BIOLOGICAL_QUALIFIER)
        cv.setBiologicalQualifierType(BQB_IS)
        cv.addResource(resource)
        metab.addCVTerm(cv)
    else:
        logging.warning(
            f"No valid IRI could be formed for {metab} with {db_id} and {DB2PREFIX_METABS.get(db_id)}:{entry}."
        )


def add_cv_term_reactions(entry: str, db_id: str, reac: Reaction):
    """Adds CVTerm to a reaction

    Args:
        - entry (str):
            Id to add as annotation
        - db_id (str):
            Database to which entry belongs. Must be in DB2PREFIX_REACS.keys().
        - reac (Reaction):
            Reaction to add CVTerm to
    """
    resource = get_identifiers_org_iri(DB2PREFIX_REACS.get(db_id), entry)
    if resource:
        if db_id == "HMDB" or db_id == "HUMAN METABOLOME DATABASE":
            if entry[:4] == "HMDB":
                entry = entry[4:]
        cv = CVTerm()
        cv.setQualifierType(BIOLOGICAL_QUALIFIER)
        cv.setBiologicalQualifierType(BQB_IS)
        cv.addResource(resource)
        reac.addCVTerm(cv)
    else:
        logging.warning(
            f"No valid IRI could be formed for {reac} with {db_id} and {DB2PREFIX_REACS.get(db_id)}:{entry}."
        )


def add_cv_term_genes(
    entry: str, db_id: str, gene: GeneProduct, lab_strain: bool = False
):
    """Adds CVTerm to a gene

    Args:
        - entry (str):
            Id to add as annotation.
        - db_id (str):
            Database to which entry belongs. Must be in DB2PREFIX_GENES.keys().
        - gene (GeneProduct):
            Gene to add CVTerm to.
        - lab_strain (bool, optional):
            For locally sequenced strains the qualifiers are always HOMOLOG_TO. Defaults to False.
    """
    resource = get_identifiers_org_iri(DB2PREFIX_GENES.get(db_id), entry)
    if resource:
        cv = CVTerm()
        cv.setQualifierType(BIOLOGICAL_QUALIFIER)
        if lab_strain:
            cv.setBiologicalQualifierType(BQB_IS_HOMOLOG_TO)
        else:
            cv.setBiologicalQualifierType(BQB_IS)
        cv.addResource(resource)
        gene.addCVTerm(cv)
    else:
        logging.warning(
            f"No valid IRI could be formed for {gene} with {db_id} and {DB2PREFIX_GENES.get(db_id)}:{entry}."
        )


def add_cv_term_pathways(entry: str, db_id: str, path: Group):
    """Add CVTerm to a groups pathway

    Args:
        - entry (str):
            Id to add as annotation
        - db_id (str):
            Database to which entry belongs. Must be in DB2PREFIX_PATHWAYS.keys().
        - path (Group):
            Pathway to add CVTerm to
    """
    resource = get_identifiers_org_iri(DB2PREFIX_PATHWAYS.get(db_id), entry)
    if resource:
        cv = CVTerm()
        cv.setQualifierType(BIOLOGICAL_QUALIFIER)
        cv.setBiologicalQualifierType(BQB_IS)
        cv.addResource(resource)
        path.addCVTerm(cv)
    else:
        logging.warning(
            f"No valid IRI could be formed for {path} with {db_id} and {DB2PREFIX_PATHWAYS.get(db_id)}:{entry}."
        )


def add_cv_term_pathways_to_entity(entry: str, db_id: str, reac: Reaction):
    """Add CVTerm to a reaction as OCCURS IN pathway

    Args:
        - entry (str):
            Id to add as annotation
        - db_id (str):
            Database to which entry belongss
        - reac (Reaction):
            Reaction to add CVTerm to
    """
    resource = get_identifiers_org_iri(DB2PREFIX_PATHWAYS.get(db_id), entry)
    if resource:
        cv = CVTerm()
        cv.setQualifierType(BIOLOGICAL_QUALIFIER)
        cv.setBiologicalQualifierType(BQB_OCCURS_IN)
        cv.addResource(resource)
        reac.addCVTerm(cv)
    else:
        logging.warning(
            f"No valid IRI could be formed for {reac} with {db_id} and {DB2PREFIX_PATHWAYS.get(db_id)}:{entry}."
        )

def get_id_from_cv_term(entity: SBase, db_id: str) -> list[str]:
    """Extract Id for a specific database from CVTerm

    Args:
        - entity (SBase):
            Species, Reaction, Gene, Pathway
        - db_id (str):
            Database of interest

    Returns:
        list[str]:
            Ids of entity belonging to db_id
    """
    def _extract_id_from_uri(uri: str) -> str:
        """Extracts the ID from the URI.

        Args:
            uri (str): 
                The URI to extract the ID from.

        Returns:
            str: 
                The extracted ID.
        """
        extracted_id = parse_iri(uri)[1]

        if not extracted_id:
            logging.info(f"Could not extract ID with bioregistry from URI: {uri}")
            extracted_id = uri.split('/')[-1]  # Fallback to splitting by "/"
            if ':' in extracted_id:
                extracted_id = extracted_id.split(':')[-1]

            if not extracted_id:
                logging.warning(f"Could not extract ID from URI: {uri} at all!")
                return None

        return extracted_id
        
    num_cvs = entity.getNumCVTerms()
    all_ids = []
    for i in range(0, num_cvs):
        ann_string = entity.getCVTerm(i)
        num_res = ann_string.getNumResources()
        ids = [
            _extract_id_from_uri(ann_string.getResourceURI(r))
            for r in range(0, num_res)
            if str(db_id) in ann_string.getResourceURI(r)
        ]
        all_ids.extend(ids)

    # Clean-up: Remove all Nones
    all_ids = [_ for _ in all_ids if _ is not None]
    if len(all_ids) == 0:
        logging.info(f'No URIs extracted for {db_id} database from {entity.getId()}')
    
    return all_ids


def generate_cvterm(qt, b_m_qt) -> CVTerm:
    """Generates a CVTerm with the provided qualifier & biological or model qualifier types

    Args:
        - qt (libSBML qualifier type):
            BIOLOGICAL_QUALIFIER or MODEL_QUALIFIER
        - b_m_qt (libSBML qualifier):
            BQM_IS, BQM_IS_HOMOLOG_TO, etc.

    Returns:
        CVTerm:
            With provided qualifier & biological or model qualifier types
    """
    cvterm = CVTerm()
    cvterm.setQualifierType(qt)

    if qt == BIOLOGICAL_QUALIFIER:
        cvterm.setBiologicalQualifierType(b_m_qt)
    else:
        cvterm.setModelQualifierType(b_m_qt)

    return cvterm


@debug
def print_cvterm(cvterm: CVTerm):
    """Debug function: Prints the URIs contained in the provided CVTerm along with the provided qualifier & biological/model qualifier types

    Args:
        cvterm (CVTerm):
            A libSBML CVTerm
    """
    if cvterm == None:
        logging.info("CVTerm currently empty!")
    else:
        current_b_m_qt = 0

        current_qt = cvterm.getQualifierType()

        if current_qt == BIOLOGICAL_QUALIFIER:
            current_b_m_qt = cvterm.getBiologicalQualifierType()
        elif current_qt == MODEL_QUALIFIER:
            current_b_m_qt = cvterm.getModelQualifierType()

        if cvterm.getNumResources() == 0:
            logging.info("No URIs present.")
        else:
            logging.info(f"Current CVTerm contains:  {cvterm.getResourceURI(0)}")

            for i in range(1, cvterm.getNumResources()):
                logging.info(f"                          {cvterm.getResourceURI(i)}")

        logging.info(
            f"Current CVTerm has QualifierType {current_qt} and Biological/ModelQualifierType {current_b_m_qt}."
        )



#!/usr/bin/env python
"""Access information from different databases or compare a model or model entities
with them. This module provides variables and function for accessing databases
for better model curation and annotation.

The following databases have functionalities implemented:

- BiGG
- ChEBI
- KEGG
- ModelSEED
- NCBI
- UniProt
"""

__author__ = """Famke Baeuerle, Gwendolyn O. Döbel, Carolin Brune, 
            Jan-Philipp Leusch and Tobias Fehrenbach"""

############################################################################
# requirements
############################################################################

import cobra
import io
import libchebipy
import numpy as np
import pandas as pd
import re
import requests
import warnings
import xmltodict

from Bio import Entrez, SeqIO
from Bio.KEGG import REST, Gene, Enzyme
from bioservices.kegg import KEGG
from cobra import Model as cobraModel
from typing import Literal, Union
from tqdm import tqdm

tqdm.pandas()
pd.options.mode.chained_assignment = None  # suppresses the pandas SettingWithCopyWarning; comment out before developing!!

from .connections import run_DIAMOND_blastp, filter_DIAMOND_blastp_results
from .io import load_a_table_from_database, create_missing_genes_protein_fasta

############################################################################
# variables
############################################################################

# database urls
# -------------
BIGG_METABOLITES_URL = "http://bigg.ucsd.edu/api/v2/universal/metabolites/"  #: :meta:

# Compartments in BiGG namespace
# ------------------------------
ALL_BIGG_COMPARTMENTS_ONE_LETTER = (
    "c",
    "e",
    "p",
    "m",
    "x",
    "r",
    "v",
    "n",
    "g",
    "u",
    "l",
    "h",
    "f",
    "s",
    "i",
    "w",
    "y",
)  #: :meta:
ALL_BIGG_COMPARTMENTS_TWO_LETTER = ("im", "cx", "um", "cm", "mm")  #: :meta:

# BioCyc databases prefixes
# -------------------------
BIOCYC_TIER1_DATABASES_PREFIXES = ["META", "ECO", "ECOLI", "HUMAN"]  #: :meta:

############################################################################
# functions
############################################################################

# BiGG
# ----
"""Map an ID to information in BiGG or compare model entities to BiGG.
"""


def get_BiGG_metabs_annot_via_dbid(
    metabolite: cobra.Metabolite, id: str, dbcol: str, compartment: str = "c"
) -> None:
    """Search for a BiGG ID and add it to a metabolite annotation.
    The search is based on a column name of the BiGG metabolite table
    and an ID to search for. Additionally, using the given compartment name,
    the found IDs are filtered for matching compartments.

    Args:
        - metabolite (cobra.Metabolite):
            The metabolite. Needs to a a COBRApy Metabolte object.
        - id (str):
            The ID to search for in the database.
        - dbcol (str):
            Name of the column of the database to check the ID against.
        - compartment (str, optional):
            The compartment name. Needs to be a valid BiGG compartment ID.
            Defaults to 'c'.
    """
    # check, if a BiGG annotation is given
    if not "bigg.metabolite" in metabolite.annotation.keys():
        # seach the database for the found id
        bigg_search = load_a_table_from_database(
            f"SELECT * FROM bigg_metabolites WHERE '{dbcol}' = '{id}'", query=True
        )
        if len(bigg_search) > 0:
            # check, if the matches also match the compartment
            metabolite.annotation["bigg.metabolite"] = [
                _.rsplit("_", 1)[0]
                for _ in bigg_search["id"].tolist()
                if _.endswith(f"_{compartment}")
            ]
            # add final matches to the annotations of the metabolite
            if len(metabolite.annotation["bigg.metabolite"]) == 0:
                metabolite.annotation.pop("bigg.metabolite")


def add_annotations_from_BiGG_metabs(metabolite: cobra.Metabolite) -> None:
    """Check a cobra.metabolite for bigg.metabolite annotations. If they exists,
    search for more annotations in the BiGG database and add them to the metabolite.

    Args:
        - metabolite (cobra.Metabolite):
            The metabolite object.
    """
    if "bigg.metabolite" in metabolite.annotation.keys():
        bigg_information = load_a_table_from_database(
            "SELECT * FROM bigg_metabolites WHERE id = '"
            + f"' OR id = '".join(metabolite.annotation["bigg.metabolite"])
            + "'",
            query=True,
        )
        db_id_bigg = {
            "BioCyc": "biocyc",
            "MetaNetX (MNX) Chemical": "metanetx.chemical",
            "SEED Compound": "seed.compound",
            "CHEBI": "chebi",
            "KEGG Compound": "kegg.compound",
            "InChI Key": "inchikey",
        }
        for db in db_id_bigg:
            info = list(set(bigg_information[db].dropna().to_list()))
            if len(info) > 0:
                info = ",".join(info)
                info = [
                    x.strip() for x in info.split(",")
                ]  # make sure all entries are a separate list object
                if db_id_bigg[db] in metabolite.annotation.keys():
                    metabolite.annotation[db_id_bigg[db]] = list(
                        set(info + metabolite.annotation[db_id_bigg[db]])
                    )
                else:
                    metabolite.annotation[db_id_bigg[db]] = info


def _add_annotations_from_bigg_reac_row(row: pd.Series, reac: cobra.Reaction) -> None:
    """Given a row of the BiGG reaction database table and a cobra.Reaction object,
    extend the annotation of the latter with the information of the former.

    Args:
        - row (pd.Series):
            The row of the database table.
        - reac (cobra.Reaction):
            The reaction object.
    """

    dbnames = {
        "RHEA": "rhea",
        "BioCyc": "biocyc",
        "MetaNetX (MNX) Equation": "metanetx.reaction",
        "EC Number": "ec-code",
        "SEED Reaction": "seed.reaction",
        "Reactome Reaction": "reactome",
        "KEGG Reaction": "kegg.reaction",
    }
    for dbname, dbprefix in dbnames.items():
        if row[dbname]:
            ids_to_add = row[dbname].strip().split(",")
            if dbprefix in reac.annotation.keys():
                reac.annotation[dbprefix] = list(
                    set(reac.annotation[dbprefix]).union(set(ids_to_add))
                )
            else:
                reac.annotation[dbprefix] = ids_to_add


# ChEBI
# -----
"""Primarily used for curating the media database. Obtain information about 
metabolites (table format)."""


def add_info_from_ChEBI_BiGG(
    missing_metabs: pd.DataFrame, charge=True, formula=True, iupac=True
) -> pd.DataFrame:
    """Adds information from CHEBI/BiGG to the provided dataframe.

    The following informations can be added:

    - charge
    - formula
    - iupac (name)

    Args:
       - missing_metabs (pd.DataFrame):
          Table containing metabolites & the respective ChEBI & BiGG IDs

    Returns:
       pd.DataFrame:
          Input table extended with the charges & chemical formulas obtained from ChEBI/BiGG.
    """

    # check if a row contains a ChEBI ID, take the first and make sure its in the format: CHEBI:234567
    def get_chebi_id(row: pd.Series) -> str:

        chebi = row.get("ChEBI")
        if pd.isnull(chebi):
            return None
        elif (
            type(chebi) == str
        ):  # check for mulitple entries (assuming first one is the best fitting one)
            chebi = chebi.split(",")[0]
            if "CHEBI" in chebi:
                return chebi
            else:
                return "CHEBI:" + chebi
        else:
            return "CHEBI:" + str(chebi)

    # Finds the charges through the ChEBI/BiGG API, defaults to: 0
    def find_charge(row: pd.Series) -> int:
        chebi_id = get_chebi_id(row)
        bigg_id = str(row.get("bigg_id"))
        charge = None
        if chebi_id:  # Get charge from ChEBI (Returns always a charge)
            chebi_entity = libchebipy.ChebiEntity(chebi_id)
            return chebi_entity.get_charge()
        elif bigg_id != "nan":  # Get charge from BiGG if no ChEBI ID available
            try:
                charge = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()[
                    "charges"
                ][
                    0
                ]  # Take first charge
            except ValueError:
                pass
            # If no charge was found, charge=0
            return charge if charge else 0

    # Finds the chemical formula through the ChEBI/BiGG API, defaults to: 'No formula'
    def find_formula(row: pd.Series) -> str:
        chebi_id = get_chebi_id(row)
        bigg_id, chem_form = str(row.get("bigg_id")), str(row.get("Chemical Formula"))
        chem_formula = None
        if chebi_id:  # Get formula from ChEBI
            chebi_entity = libchebipy.ChebiEntity(chebi_id)
            chem_formula = chebi_entity.get_formula()
        if not chem_formula:  # If no formula was found with ChEBI/No ChEBI ID available
            if bigg_id != "nan":  # Get formula from BiGG
                try:
                    chem_formula = requests.get(
                        BIGG_METABOLITES_URL + bigg_id[:-2]
                    ).json()["formulae"][
                        0
                    ]  # Take first formula
                except ValueError:
                    pass
            if not chem_formula:  # If no formula was found with BiGG ID
                # Get formula already existing in dataframe or set to 'No formula'
                chem_formula = chem_form if chem_form != "nan" else "No formula"
        return chem_formula

    # using the ChEBI ID, retrieve the IUPAC name from the ChEBI database.
    def find_iupac(row: pd.Series) -> str:

        chebi_id = get_chebi_id(row)
        if chebi_id:
            chebi_entity = libchebipy.ChebiEntity(chebi_id)
            # only take the IUPAC names
            iupac_names = []
            for name in chebi_entity.get_names():
                print(name)
                if name.get_source() == "IUPAC":
                    iupac_names.append(name.get_name())
                    break

            if len(iupac_names) > 0:
                iupac_name = ", ".join(iupac_names)
            else:
                iupac_name = None
        else:
            iupac_name = None

        return iupac_name

    if charge:
        missing_metabs["charge"] = missing_metabs.apply(find_charge, axis=1)
    if formula:
        missing_metabs["New Chemical Formula"] = missing_metabs.apply(
            find_formula, axis=1
        )
        missing_metabs["Chemical Formula"] = missing_metabs["New Chemical Formula"]
        missing_metabs.drop("New Chemical Formula", axis=1, inplace=True)
    if iupac:
        missing_metabs["ChEBI_IUPAC"] = missing_metabs.apply(find_iupac, axis=1)

    return missing_metabs


# KEGG
# ----
"""Retrieve and parse information from the KEGG database. 
"""


def get_kegg_genes(organismid: str) -> pd.DataFrame:
    """Extracts list of genes from KEGG given an organism

    Args:
        - organismid (str):
            KEGG ID of organism which the model is based on

    Returns:
        pd.DataFrame:
            Table of all genes denoted in KEGG for the organism
    """
    k = KEGG()
    gene_list = k.list(organismid)

    return pd.read_table(io.StringIO(gene_list), header=None)


def parse_KEGG_gene(locus_tag: str) -> dict:
    """Based on a locus tag, fetch the corresponding KEGG entry and
    parse it into a dictionary containing the following information (if available):

    - ec-code
    - orthology
    - references

    Args:
        - locus_tag (str):
            The locus in the format <organism_id>:<locus_tag>

    Returns:
        dict:
            The collected information.
    """

    gene_info = dict()
    gene_info["orgid:locus"] = locus_tag

    # retireve KEGG gene entry
    try:
        gene_entry = list(Gene.parse(REST.kegg_get(locus_tag)))[0]
    except Exception as e:
        gene_entry = None

    # skip, if no entry found
    if not gene_entry:
        gene_info["ec-code"] = None
        return gene_info

    # extract orthology and ec-code
    if len(gene_entry.orthology) > 0:
        # get KEGG orthology ID
        kegg_orthology = [_[0] for _ in gene_entry.orthology]
        gene_info["kegg.orthology"] = kegg_orthology
        # get EC number
        ec_numbers = [
            re.search(r"(?<=EC:).*(?=\])", orth[1]).group(0)
            for orth in gene_entry.orthology
            if re.search(r"(?<=EC:).*(?=\])", orth[1])
        ]
        if isinstance(ec_numbers, list) and len(ec_numbers) > 0:
            gene_info["ec-code"] = [
                ec for ec_str in ec_numbers for ec in ec_str.split(" ")
            ]

    if not "ec-code" in gene_info.keys():
        gene_info["ec-code"] = None

    # get more information about connections to other databases
    if len(gene_entry.dblinks) > 0:
        for dbname, ids in gene_entry.dblinks:
            conform_dbname = re.sub(
                pattern=r"(NCBI)(.*)(ID$)", repl="\\1\\2", string=dbname
            )  # Remove ID if NCBI in name
            conform_dbname = re.sub(
                r"[^\w]", "", conform_dbname
            )  # remove special signs except underscore
            conform_dbname = conform_dbname.lower()  # make lower case
            gene_info[conform_dbname] = ids

    return gene_info


def parse_KEGG_ec(ec: str) -> dict:
    """Based on an EC number, fetch the corresponding KEGG entry and
    parse it into a dictionary containing the following information (if available):

    - ec-code
    - id (kegg.reference)
    - equation
    - reference
    - pathway

    Args:
        - ec (str):
            The EC number in the format 'x.x.x.x'

    Returns:
        dict:
            The collected information about the KEGG entry.
    """

    ec_info = dict()
    ec_info["ec-code"] = ec

    # retrieve KEGG entry
    try:
        ec_entry = list(Enzyme.parse(REST.kegg_get(ec)))[0]
    except Exception as e:
        ec_entry = None
        ec_info["id"] = None
        ec_info["equation"] = None
        ec_info["reference"] = None
        return ec_info

    # retrieve reaction information from entry
    rn_numbers = [
        re.search(r"(?<=RN:).*(?=\])", reac).group(0)
        for reac in ec_entry.reaction
        if re.search(r"(?<=RN:).*(?=\])", reac)
    ]
    if len(rn_numbers) > 0:
        ec_info["id"] = [_.split(" ") for _ in rn_numbers]
        ec_info["equation"] = None
    else:
        ec_info["id"] = None
        ec_info["equation"] = ec_entry.reaction

    # retrieve database links from entry
    refs = dict()
    # orthology not possible with biopython
    if len(ec_entry.pathway) > 0:
        refs["kegg.pathway"] = [_[1] for _ in ec_entry.pathway]
    if len(ec_entry.dblinks) > 0:
        for dbname, ids in ec_entry.dblinks:
            if "BRENDA" in dbname:
                refs["brenda"] = ids
            if "CAS" == dbname:
                refs["cas"] = ids
    ec_info["reference"] = refs

    return ec_info


def kegg_reaction_parser(rn_id: str) -> dict:
    """Get the entry of a KEGG reaction ID and
    parse the information into a dictionary.

    Args:
        - rn_id (str):
            A reaction ID existing in KEGG.

    Returns:
        dict:
            The KEGG entry information as a dictionary.
    """

    # get KEGG reaction entry
    try:
        kegg_reac = REST.kegg_get(f"rn:{rn_id}")
        kegg_reac = kegg_reac.read()
    except Exception as e:
        return None

    # parse the entry for necessary information
    features = {"db": []}
    collect = False
    for line in kegg_reac.split("\n"):
        if line:
            if line.startswith("NAME"):
                features["name"] = line.replace("NAME", "", 1).strip()
            elif line.startswith("EQUATION"):
                features["equation"] = line.replace("EQUATION", "", 1).strip()
            elif line.startswith("ENZYME"):
                features["ec-code"] = re.split(
                    r"\s+", line.replace("ENZYME", "", 1).strip()
                )
                collect = "ec-code"
            elif line.startswith("RCLASS"):
                features["kegg.rclass"] = [
                    line.replace("RCLASS", "", 1).strip().split(" ")[0]
                ]
                collect = "kegg.rclass"
            elif line.startswith("PATHWAY"):
                features["kegg.pathway"] = [
                    line.replace("PATHWAY", "", 1).strip().split(" ")[0]
                ]
                collect = "kegg.pathway"
            elif line.startswith("DBLINKS"):
                features["db"] = [line.replace("DBLINKS", "", 1).strip()]
                collect = "db"
            elif collect and line[0] != "/":
                if len(features["db"]) == 0:
                    if line[0].isupper():
                        collect = None
                    else:
                        line = line.strip()
                        match collect:
                            case "ec-code":
                                features[collect].append(line)
                            case _:
                                features[collect].append(line.split(" ")[0])
                else:
                    features["db"].append(line.strip())
            else:
                continue

    # parse references
    db_entries = features["db"]
    features["db"] = {"kegg.reaction": rn_id}
    for entry in db_entries:
        db, identifier = entry.split(":")
        db = db.strip().lower()
        if db in features["db"]:
            features["db"][db] = features["db"][db].append(identifier)
        else:
            features["db"][db] = [identifier.strip()]

    if "ec-code" in features:
        features["db"]["ec-code"] = features["ec-code"]
        del features["ec-code"]
    if "kegg.rclass" in features:
        features["db"]["kegg.rclass"] = features["kegg.rclass"]
        del features["kegg.rclass"]
    if "kegg.pathway" in features:
        features["db"]["kegg.pathway"] = features["kegg.pathway"]
        del features["kegg.pathway"]

    return features


# ModelSEED
# ---------
""" Reports mismatches in charges and formulae based on ModelSEED

Extracts ModelSEED data from a given tsv file, extracts all metabolites from a given model. Both lists of metabolites are compared by charge and formula.
"""


def get_modelseed_compounds() -> pd.DataFrame:
    """Extracts compounds from ModelSEED which have BiGG Ids

    Returns:
        pd.DataFrame:
            Table containing ModelSEED data
    """
    # Get only rows where BiGG is contained
    com = load_a_table_from_database(
        "SELECT *, INSTR(aliases, 'BiGG:') bigg FROM modelseed_compounds WHERE bigg > 0"
    )

    def get_bigg_ids(aliases):
        try:
            aliases_list = aliases.split("|")
            bigg = [x[6:] for x in aliases_list if re.search(r"BiGG: .*", x)]
            return bigg[0]
        except (IndexError, AttributeError):
            return None

    com["BiGG"] = com.apply(lambda row: get_bigg_ids(row["aliases"]), axis=1)

    return com.loc[:, ["id", "name", "formula", "mass", "charge", "BiGG"]]


def get_model_charges(model: cobraModel) -> pd.DataFrame:
    """Extracts all metabolites from model

    Args:
        - model (cobraModel):
            Model loaded with COBRApy

    Returns:
        pd.DataFrame:
            Table containing charges and formulae of model metabolites
    """
    charges = {}
    for metab in model.metabolites:
        charges[metab.id[:-2]] = [metab.charge, metab.formula]

    df_charges = (
        pd.DataFrame.from_dict(
            charges, orient="index", columns=["charge_model", "formula_model"]
        )
        .reset_index()
        .rename(columns={"index": "BiGG"})
    )

    return df_charges


def get_modelseed_charges(modelseed_compounds: pd.DataFrame) -> pd.DataFrame:
    """Extract table with BiGG, charges and formulae

    Args:
        - modelseed_compounds (pd.DataFrame):
            ModelSEED data. Output from :py:func:`~refinegems.utility.db_access.get_modelseed_compounds`.

    Returns:
        pd.DataFrame:
            Table containing charges and formulae of ModelSEED metabolites
    """
    modelseed_compounds = modelseed_compounds.loc[
        :, ["charge", "BiGG", "formula"]
    ].rename(columns={"charge": "charge_modelseed", "formula": "formula_modelseed"})
    lst_col = "BiGG"
    x = modelseed_compounds.assign(
        **{lst_col: modelseed_compounds[lst_col].str.split(";")}
    )
    df_ms = pd.DataFrame(
        {
            col: np.repeat(x[col].values, x[lst_col].str.len())
            for col in x.columns.difference([lst_col])
        }
    ).assign(**{lst_col: np.concatenate(x[lst_col].values)})[x.columns.tolist()]
    return df_ms


def compare_model_modelseed(
    model_charges: pd.DataFrame, modelseed_charges: pd.DataFrame
) -> pd.DataFrame:
    """Compares tables with charges / formulae from model & modelseed

    Args:
        - model_charges (pd.DataFrame):
            Charges and formulae of model metabolites. Output of :py:func:`~refinegems.utility.db_access.get_model_charges`.
        - modelseed_charges (pd.DataFrame):
            Charges and formulae of ModelSEED metabolites. Output of :py:func:`~refinegems.utility.db_access.get_modelseed_charges`.

    Returns:
        pd.DataFrame:
            Table containing info whether charges / formulae match
    """
    df_comp = pd.merge(model_charges, modelseed_charges, on="BiGG", how="left")

    def f(x):
        return True if float(x["charge_model"]) == x["charge_modelseed"] else False

    def g(x):
        return True if x["formula_model"] == x["formula_modelseed"] else False

    df_comp["charge_match"] = df_comp.apply(f, axis=1)
    df_comp["formula_match"] = df_comp.apply(g, axis=1)

    return df_comp


def get_charge_mismatch(df_comp: pd.DataFrame) -> pd.DataFrame:
    """Extracts metabolites with charge mismatch of model & modelseed

    Args:
        df_comp (pd.DataFrame):
            Charge and formula mismatches. Output from :py:func:`~refinegems.utility.db_access.compare_model_modelseed`.

    Returns:
        pd.DataFrame:
            Table containing metabolites with charge mismatch
    """
    return df_comp.loc[~df_comp["charge_match"]].dropna(subset=["charge_modelseed"])


def get_formula_mismatch(df_comp: pd.DataFrame) -> pd.DataFrame:
    """Extracts metabolites with formula mismatch of model & modelseed

    Args:
        df_comp (pd.DataFrame):
            Charge and formula mismatches. Output from :py:func:`~refinegems.utility.db_access.compare_model_modelseed`.

    Returns:
        pd.DataFrame:
            Table containing metabolites with formula mismatch
    """
    return df_comp.loc[~df_comp["formula_match"]].dropna(subset=["formula_modelseed"])


def get_compared_formulae(formula_mismatch: pd.DataFrame) -> pd.DataFrame:
    """Compare formula by atom pattern

    Args:
        formula_mismatch (pd.DataFrame):
            Table with column containing atom comparison. Output from :py:func:`~refinegems.utility.db_access.get_formula_mismatch`.

    Returns:
        pd.DataFrame:
            table containing metabolites with formula mismatch
    """

    def formula_comparison(f1, f2):
        # from Jan Leusch
        formula_pattern = r"[A-Z][a-z]?\\d*"
        atom_pattern = r"[A-Z][a-z]?"
        atom_number_pattern = r"\\d+"
        difference = {}
        f1_dict = {}
        f2_dict = {}
        f1 = re.findall(formula_pattern, f1)
        f2 = re.findall(formula_pattern, f2)
        for p in f1:
            key = re.findall(atom_pattern, p)[0]
            value = re.findall(atom_number_pattern, p)
            if not value:
                value = 1
            else:
                value = (int)(value[0])
            f1_dict[key] = value
        for q in f2:
            key = re.findall(atom_pattern, q)[0]
            value = re.findall(atom_number_pattern, q)
            if not value:
                value = 1
            else:
                value = (int)(value[0])
            f2_dict[key] = value
        difference = f1_dict
        if not f2_dict:
            return difference
        else:
            for q in f2_dict:
                if q in difference:
                    difference[q] -= f2_dict[q]
                else:
                    difference[q] = -f2_dict[q]
        for a in list(difference):
            if difference[a] == 0:
                difference.pop(a)
        return difference

    formula_mismatch["formula_comparison"] = formula_mismatch.apply(
        lambda row: formula_comparison(row["formula_model"], row["formula_modelseed"]),
        axis=1,
    )

    return formula_mismatch


def compare_to_modelseed(model: cobraModel) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Executes all steps to compare model metabolites to ModelSEED metabolites

    Args:
        - model (cobraModel):
            Model loaded with COBRApy

    Returns:
        tuple:
            Tables with charge (1) & formula (2) mismatches

            (1) pd.DataFrame: Table with charge mismatches
            (2) pd.DataFrame: Table with formula mismatches
    """
    ms_comp = get_modelseed_compounds()
    model_charges = get_model_charges(model)
    modelseed_charges = get_modelseed_charges(ms_comp)
    df_comp = compare_model_modelseed(model_charges, modelseed_charges)
    charge_mismatch = get_charge_mismatch(df_comp)
    formula_mismatch = get_formula_mismatch(df_comp)
    formula_comp = get_compared_formulae(formula_mismatch)
    return charge_mismatch, formula_comp


# NCBI
# ----
"""Connect to NBCI (using Biopython/Entrez) to extract information
like EC number of an NBCI protein accession number or information about locus tags.
"""


def _search_ncbi_for_gp(
    row: pd.Series, id_type: Literal["refseq", "ncbiprotein"]
) -> pd.Series:
    """Fetches protein name and locus tag from NCBI

    Args:
        - row (pd.Series):
            Row of a pandas DataFrame containing RefSeq/NCBI Protein IDs in columns
        - id_type (Literal['refseq', 'ncbiprotein']):
            ID type of IDs in provided row.
            Can be one of ['refseq', 'ncbiprotein'].

    Returns:
        pd.Series:
            Modified input row
    """
    match id_type:
        case "refseq":
            ncbi_id = row["REFSEQ"]
        case "ncbiprotein":
            ncbi_id = row["NCBI"]
        case _:
            raise TypeError(
                f"Provided type for id_type {id_type} invalid. Please use one of ['refseq', 'ncbiprotein']."
            )
            return row

    if not pd.isnull(ncbi_id):

        try:
            handle = Entrez.efetch(
                db="protein", id=ncbi_id, rettype="gbwithparts", retmode="text"
            )
            records = SeqIO.parse(handle, "gb")

            for i, record in enumerate(records):
                match id_type:
                    case "refseq":
                        row["name_refseq"] = record.description
                    case "ncbiprotein":
                        for feature in record.features:
                            if feature.type == "CDS":
                                row["name_ncbi"] = record.description
                                row["locus_tag"] = feature.qualifiers["locus_tag"][0]

        except Exception as e:
            print(f"{e} with {id_type} ID {ncbi_id}")

    return row


# fetching the EC number (if possible) from NCBI
# based on an NCBI protein ID (accession version number)
def get_ec_from_ncbi(mail: str, ncbiprot: str) -> Union[str, None]:
    """Based on a NCBI protein accession number, try and fetch the
    EC number from NCBI.

    Args:
        - mail (str):
            User's mail address for the NCBI ENtrez tool.
        - ncbiprot (str):
            The NCBI protein accession number.

    Returns:
        (1) Case: fetching successful
                str:
                    The EC number associated with the protein ID based on NCBI.

        (2) Case: fetching unsuccessful
                None:
                    Nothing to return
    """
    try:
        Entrez.email = mail
        handle = Entrez.efetch(db="protein", id=ncbiprot, rettype="gpc", retmode="xml")
        for feature_dict in xmltodict.parse(handle)["INSDSet"]["INSDSeq"][
            "INSDSeq_feature-table"
        ]["INSDFeature"]:
            if isinstance(feature_dict["INSDFeature_quals"]["INSDQualifier"], list):
                for qual in feature_dict["INSDFeature_quals"]["INSDQualifier"]:
                    if qual["INSDQualifier_name"] == "EC_number":
                        return qual["INSDQualifier_value"]
    except Exception as e:
        return None


# Uniprot
# -------
"""Currently mainly used for gapfilling, map your proteins against EC numbers 
using the SwissProt database"""


def map_dmnd_res_to_sp_ec_brenda(
    dmnd_results: pd.DataFrame, swissprot_mapping_path: str
) -> pd.DataFrame:
    """Map the results of a DIAMOND BLASTp run (filtered, see
    :py:func:`~refinegems.curation.db_access.db.filter_DIAMOND_blastp_results`)

    Args:
        - dmnd_results (pd.DataFrame):
            The results of the DIAMOND run.
        - swissprot_mapping_path (str):
            The path to the SwissProt mapping file (IDs against BRENDA and EC,
            for information on how to get them, refer to :py:func:`~refinegems.utility.set_up.download_url`)

    Returns:
        pd.DataFrame:
            The resulting mapping (no duplicates).
    """

    def _combine_EC_BRENDA(brenda: str, ec: str) -> list:
        """Helper function for combining the entries for BRENDA and EC number
        for one ID (SwissProt mapping file) into a list of EC numbers.

        Args:
            - brenda (str):
                The BRENDA column value.
            - ec (str):
                The EC number column value.

        Returns:
            list:
                The list with the EC numbers.
        """

        nums_list = []
        # get brenda ids
        if brenda and not pd.isna(brenda):
            for num in brenda.split(";"):
                nums_list.append(num.strip())

        # get ec numbers
        if ec and not pd.isna(ec):
            for num in ec.split(";"):
                nums_list.append(num.strip())

        # remove duplicates
        nums_list = list(set(nums_list))
        # filter non-complete EC numbers : X.X.X.X => 7 or more characters
        nums_list = [_ for _ in nums_list if len(_) >= 7]

        return nums_list

    # load the SwissProt mapping file
    swissprot_mapping = pd.read_csv(swissprot_mapping_path, sep="\t")
    swissprot_mapping.dropna(subset=["BRENDA", "EC number"], how="all", inplace=True)

    # extract SwissProts IDs from subject_ID
    dmnd_results.columns = ["locus_tag", "UniProt"]
    dmnd_results["UniProt"] = dmnd_results["UniProt"].apply(lambda x: x.split("|")[1])

    # match
    dmnd_results = dmnd_results.merge(
        swissprot_mapping, left_on="UniProt", right_on="Entry", how="left"
    )
    dmnd_results.drop("Entry", axis=1, inplace=True)
    dmnd_results["ec-code"] = dmnd_results.apply(
        lambda x: _combine_EC_BRENDA(x["BRENDA"], x["EC number"]), axis=1
    )
    dmnd_results.drop(["BRENDA", "EC number"], axis=1, inplace=True)
    dmnd_results = dmnd_results.explode("ec-code")
    dmnd_results.drop_duplicates(inplace=True)

    return dmnd_results


def get_ec_via_swissprot(
    fasta: str,
    db: str,
    missing_genes: pd.DataFrame,
    swissprot_mapping_file: str,
    outdir: str = None,
    sens: Literal[
        "sensitive", "more-sensitive", "very-sensitive", "ultra-sensitive"
    ] = "more-sensitive",
    cov: float = 95.0,
    t: int = 2,
    pid: float = 90.0,
) -> pd.DataFrame:
    """Based on a protein FASTA and a missing genes tables, mapped them to EC numbers
    using a Swissprot DIAMOND database and a SwissProt mapping file (see :py:func:`~refinegems.utility.set_up.download_url`
    on how to download the needed files).

    Args:
        - fasta (str):
            Path to the FASTA protein file.
        - db (str):
            Path to the DIAMOND database (SwissProt).
        - missing_genes (pd.DataFrame):
            The table of missing genes.
        - swissprot_mapping_file (str):
            Path to the SwissProt mapping file.
        - outdir (str, optional):
            Path to a directory to write the output to.
            Defaults to None.
        - sens (Literal['sensitive', 'more-sensitive', 'very-sensitive','ultra-sensitive'], optional):
            Sensitivity mode of DIAMOND blastp. Defaults to 'more-sensitive'.
        - cov (float, optional):
            Coverage threshold for DIAMOND blastp. Defaults to 95.0.
        - t (int, optional):
            Number of threads to use for DIAMOND blastp. Defaults to 2.
        - pid (float, optional):
            Percentage identity value to use as a cutoff for the results
            of the DIAMOND blastp run. Defaults to 90.0.

    Returns:
        pd.DataFrame:
            The missing genes table extended by the mapping to an EC number,
            if successful.
    """

    # Step 1: Make a FASTA out of the missing genes
    miss_fasta = create_missing_genes_protein_fasta(fasta, missing_genes, outdir)
    # Step 2: Run DIAMOND
    #         blastp mode against SwissProt DB
    blast_path = run_DIAMOND_blastp(
        miss_fasta, db, sensitivity=sens, coverage=cov, threads=t, outdir=outdir
    )
    # Step 3: filter DIAMOND hits
    dmnd_res = filter_DIAMOND_blastp_results(blast_path, pid)
    # Step 4: map to Swissprot mapping file
    mapped_res = map_dmnd_res_to_sp_ec_brenda(dmnd_res, swissprot_mapping_file)
    # Step 5: Aggregate UniProt IDs for unique combinations of
    #         EC numbers and locus tags
    mapped_res = (
        mapped_res.groupby(["locus_tag", "ec-code"])
        .agg({"UniProt": lambda x: x.tolist()})
        .reset_index()
    )

    return mapped_res


# DIAMOND database
# ----------------
"""Handle and retrieve information from DIAMOND databases"""


def map_to_homologs(
    fasta: str,
    db: str,
    missing_genes: pd.DataFrame,
    mapping_file: str,
    outdir: str = None,
    sens: Literal[
        "sensitive", "more-sensitive", "very-sensitive", "ultra-sensitive"
    ] = "more-sensitive",
    cov: float = 95.0,
    t: int = 2,
    pid: float = 90.0,
    email=None,
) -> pd.DataFrame:

    # Step 1: Make a FASTA out of the missing genes
    miss_fasta = create_missing_genes_protein_fasta(fasta, missing_genes, outdir)

    # Step 2: Run DIAMOND
    #         blastp mode against SwissProt DB
    blast_path = run_DIAMOND_blastp(
        miss_fasta, db, sensitivity=sens, coverage=cov, threads=t, outdir=outdir
    )
    # Step 3: filter DIAMOND hits
    dmnd_res = filter_DIAMOND_blastp_results(blast_path, pid)
    dmnd_res.rename(
        columns={"query_ID": "locus_tag", "subject_ID": "ncbiprotein"}, inplace=True
    )

    # Step 4: map to NCBI information
    # if a precomputet mapping exists
    if mapping_file:
        # load pre-compiles ncbi mapping
        ncbi_mapping = pd.read_csv(mapping_file, dtype="str")

        # merge with input info table:
        table = pd.merge(
            dmnd_res,
            ncbi_mapping,
            left_on="ncbiprotein",
            right_on="ncbi_accession_version",
        )
        table.replace({"-": None}, inplace=True)
        table.drop(columns=["ncbi_accession_version"], inplace=True, axis=1)
        table.rename(columns={"EC number": "ec-code"}, inplace=True)

        # aggregate to reduce redundant information
        table = (
            table.groupby(["locus_tag", "ec-code"], dropna=False)
            .agg(
                {
                    "ncbiprotein": "first",  # lambda x: [_ for _ in x if not pd.isna(_)]
                    "locus_tag_ref": "first",  # lambda x: [_ for _ in x if not pd.isna(_)]
                    "old_locus_tag": "first",  # lambda x: [_ for _ in x if not pd.isna(_)]
                    "GeneID": "first",  # lambda x: [_ for _ in x if not pd.isna(_)]
                }
            )
            .reset_index()
        )
        # locus_tag ncbiprotein locus_tag_ref old_locus_tag GeneID EC number

    # no mapping file
    else:
        if email:
            print(
                "\t\tNo precomputed mapping. Retrieving information directly from NCBI.\n\t\tThis may take a while."
            )
            table["locus_tag_ref"] = pd.Series(dtype="str")
            table["old_locus_tag"] = pd.Series(dtype="str")
            table["GeneID"] = pd.Series(dtype="str")
            # .......................
            # @DEBUG
            # if len(table) > 10:
            #     table = table.sample(10)
            #     print('Running in debugging mode')
            # .......................
            table["ec-code"] = table["ncbiprotein"].progress_apply(
                lambda x: get_ec_from_ncbi(x, email), axis=1
            )
        else:
            warnings.warn(
                "No email provided for NCBI quieries. Skipping mapping to EC numbers."
            )

    return table



#!/usr/bin/env python
"""This module provides functions for adding, handling and analysing
the KEGG pathways (or more specific their annotations) contained in a model.
"""

__author__ = "Famke Baeuerle and Carolin Brune"

############################################################################
# requirements
############################################################################

import cobra
import re
import urllib

from bioservices import KEGG
from Bio.KEGG import REST, Enzyme
from libsbml import SBMLReader, GroupsExtension
from libsbml import Model as libModel
from tqdm.auto import tqdm

from ..utility.cvterms import (
    add_cv_term_pathways,
    get_id_from_cv_term,
    add_cv_term_pathways_to_entity,
)
from ..utility.db_access import kegg_reaction_parser
from ..classes.reports import KEGGPathwayAnalysisReport

############################################################################
# functions
############################################################################

# adding KEGG reactions as Group Pathways
# ---------------------------------------
""" 
If your organism occurs in the KEGG database, 
extract the KEGG reaction ID from the annotations of your reactions and identify, 
in which KEGG pathways this reaction occurs. 
Add all KEGG pathways for a reaction then as annotations 
with the biological qualifier `OCCURS_IN` to the respective reaction.
"""


def load_model_enable_groups(modelpath: str) -> libModel:
    """Loads model as document using libSBML and enables groups extension

    Args:
        - modelpath (str):
            Path to GEM

    Returns:
        libModel:
            Model loaded with libSBML
    """
    reader = SBMLReader()
    read = reader.readSBMLFromFile(modelpath)  # read from file
    groupextension = GroupsExtension()
    groupURI = groupextension.getURI(3, 1, 1)
    read.enablePackage(groupURI, "groups", True)  # enable groups extension
    read.setPkgRequired("groups", False)  # make groups not required
    model = read.getModel()
    return model


def _extract_kegg_ec_from_reac(model: libModel) -> tuple[dict, list]:
    """Given a model. extract KEGG IDs and EC numbers from all reactions.

    Args:
        - model (libModel):
            A model loaded with libSBML.

    Returns:
        tuple[dict, list]:
            Tuple with (1) a dictionary with the mappings and (2) a list of unmapped reactions.

            (1) Dictionary:
                Reaction Id as key and dictionary as value.
                The dictionary contains either one or both of the following
                keys with the corresponding values:

                - 'kegg.reaction': List of KEGG reaction Ids
                - 'eccode': List of EC numbers

            (2) List:
                Reaction Ids without KEGG or EC annotation.
    """

    list_reac = model.getListOfReactions()
    mapped = {}
    non_mapped = []

    for reaction in list_reac:
        ecs = get_id_from_cv_term(reaction, "ec-code")
        kegg_ids = get_id_from_cv_term(reaction, "kegg.reaction")
        mappings = dict()
        if len(kegg_ids) > 0:
            mappings["kegg.reaction"] = kegg_ids
        if len(ecs) > 0:
            mappings["eccode"] = ecs
        if len(mappings) == 0:
            non_mapped.append(reaction.getId())
        else:
            mapped[reaction.getId()] = mappings

    # mapped : dict = {'reac_id': {kegg: [], ec-code:[]}}
    return mapped, non_mapped


def find_kegg_pathways(
    mapped_reacs: dict, viaEC: bool = False, viaRC: bool = False
) -> dict:
    """Given a dictionary of reaction IDs mapped to KEGG reaction IDs and/or EC numbers,
    extract the KEGG pathways for each reaction based on the KEGG reaction ID.

    Args:
        - mapped_reacs (dict):
            Dictionary containing the information about the reactions.
            For more information see, :py:func:`~refinegems.curation.pathways._extract_kegg_ec_from_reac`.
        - viaEC (bool, optional):
            If True, also tries mapping to pathways via EC number, if
            via reaction ID is unsuccessful.
            Defaults to False.
        - viaRC (bool, optional):
            If True, also tries mapping to pathways via reaction class, if
            via reaction ID is unsuccessful.
            Defaults to False.

    Returns:
        dict:
            Dictionary with the reaction IDs as keys and a list of KEGG pathway IDs as values.
    """

    def _get_pathway_via_rc(rc_list: list[str]) -> list[str]:
        """Helper function to extract KEGG pathways IDs by a list of reaction class IDs.

        Args:
            - rc_list (list[str]):
                List of reaction class IDs.

        Returns:
            list[str]:
                List of extracted pathways.
                If none are found, return an empty list.
        """

        pathway_ids = []
        collect = False
        for kegg_rc in rc_list:
            kegg_rc = REST.kegg_get(kegg_rc)
            kegg_rc = kegg_rc.read()
            for line in kegg_rc.split("\n"):
                if line:
                    if line.startswith("PATHWAY"):
                        collect = True
                        pathway_ids.append(
                            line.replace("PATHWAY", "", 1).strip().split(" ")[0]
                        )
                    elif collect == True and line[0] != "/":
                        if line[0].isupper():
                            collect = False
                        else:
                            pathway_ids.append(line.strip().split(" ")[0])

        return pathway_ids

    kegg_pathways = {}

    print("Extracting pathway Id for each reaction:")
    for reac_id in tqdm(mapped_reacs.keys(), unit="reaction"):

        kegg_reaction = None
        pathways = []

        # via reaction
        if "kegg.reaction" in mapped_reacs[reac_id].keys():
            for kegg_id in mapped_reacs[reac_id]["kegg.reaction"]:
                try:
                    kegg_reaction = kegg_reaction_parser(kegg_id)
                    if isinstance(kegg_reaction["db"]["kegg.pathway"], list):
                        pathways.extend(kegg_reaction["db"]["kegg.pathway"])
                    else:
                        pathways.append(kegg_reaction["db"]["kegg.pathway"])

                # exception handling
                except urllib.error.HTTPError:
                    print(f"HTTPError: {reac_id} on {kegg_id}")
                except ConnectionResetError:
                    print(f"ConnectionResetError: {reac_id} on {kegg_id}")
                except urllib.error.URLError:
                    print(f"URLError: {reac_id} on {kegg_id}")
                except KeyError:
                    # no pathway found
                    pass
                # except Exception as e:
                #     print(F'Something unexpected happened: {reac_id} on {kegg_id}')
                #     print(repr(e))

        # via RC
        # via reaction class
        # can lead to some additional classes, as the RC are not as strictly defined as
        # the reactions themselves

        if viaRC and kegg_reaction and len(pathways) == 0:

            try:
                rc_results = _get_pathway_via_rc(kegg_reaction["db"]["kegg.rclass"])
                if len(rc_results) > 0:
                    pathways.extend(rc_results)
            # exception handling
            except urllib.error.HTTPError:
                print(f"HTTPError: {reac_id} on RCLASS")
            except ConnectionResetError:
                print(f"ConnectionResetError: {reac_id} on RCLASS")
            except urllib.error.URLError:
                print(f"URLError: {reac_id} on RCLASS")
            except KeyError:
                # no pathway found
                pass
            except Exception as e:
                print(f"Something unexpected happened: {reac_id} on RCLASS")

        # via EC
        # seems really sketchy to do it this way, as ONE EC number
        # can include MANY reactions for different pathways

        if viaEC and len(pathways) == 0:
            ec = None
            # tests, if reaction already has an annotated EC number
            if "eccode" in mapped_reacs[reac_id].keys():
                ec = mapped_reacs[reac_id]["eccode"]
            # try to get via kegg reaction
            elif (
                kegg_reaction
                and "db" in kegg_reaction.keys()
                and "ec-code" in kegg_reaction["db"]
            ):
                ec = kegg_reaction["db"]["ec-code"]

            # if EC number found,
            # get pathways information using the EC number
            if ec:
                ecnum = None
                try:
                    for ecnum in ec:
                        kegg_ec = REST.kegg_get(f"ec:{ecnum}")
                        kegg_ec = Enzyme.read(kegg_ec)
                        # no pathway found
                        if len(kegg_ec.pathway) == 0 or kegg_ec.pathway == None:
                            pass
                        # pathway found
                        else:
                            for i in kegg_ec.pathway:
                                pathways.append(i[1])
                # exception handling
                except urllib.error.HTTPError:
                    print(f"HTTPError: {reac_id} on {ecnum}")
                except ConnectionResetError:
                    print(f"ConnectionResetError: {reac_id} on {ecnum}")
                except urllib.error.URLError:
                    print(f"URLError: {reac_id} on {ecnum}")
                except KeyError:
                    # no pathway found
                    pass
                except Exception as e:
                    print(f"Something unexpected happened: {reac_id} on {ecnum}")

        # store the pathways
        kegg_pathways[reac_id] = list(set(pathways))

    return kegg_pathways


def add_kegg_pathways(model, kegg_pathways) -> libModel:
    """Add KEGG reactions as BQB_OCCURS_IN.

    Args:
        - model (libModel):
            Model loaded with libSBML. Output of :py:func:`~refinegems.curation.pathways.load_model_enable_groups`.
        - kegg_pathways (dict):
            Reaction Id as key and KEGG Pathway Id as value, e.g. see output of :py:func:`~refinegems.curation.pathways.find_kegg_pathways`.

    Returns:
        libModel:
        Modified model with KEGG pathways.
    """
    list_reac = model.getListOfReactions()

    for reaction in list_reac:
        if reaction.getId() in kegg_pathways.keys():
            for path in kegg_pathways[reaction.getId()]:
                add_cv_term_pathways_to_entity(path, "KEGG", reaction)

    return model


def create_pathway_groups(model: libModel, pathway_groups) -> libModel:
    """Use group module to add reactions to KEGG pathway.

    Args:
        - model (libModel):
            Model loaded with libSBML. Output of :py:func:`~refinegems.curation.pathways.load_model_enable_groups`.
        - pathway_groups (dict):
            KEGG Pathway Id as key and reactions Ids as values, e.g. see output of :py:func:`~refinegmes.curation.pathways.set_kegg_pathways._invert_reac_pathway_dict`.

    Returns:
        libModel:
            Modified model with groups for pathways.
    """
    k = KEGG()
    groups = model.getPlugin("groups")
    group_list = groups.getListOfGroups()
    keys = list(pathway_groups.keys())

    print("Adding pathways as groups to the model:")
    for i in tqdm(range(len(pathway_groups))):
        kegg_pathway = k.get(keys[i])
        dbentry = k.parse(kegg_pathway)
        if groups.getGroup("G_" + keys[i]) is not None:
            group = groups.getGroup("G_" + keys[i])
            group.setName(dbentry["NAME"][0])
            group.setMetaId("meta_" + "G_" + keys[i])
            group.setKind("partonomy")
            group.setSBOTerm("SBO:0000633")  # NAME
            add_cv_term_pathways(keys[i], "KEGG", group)
            for reac in pathway_groups[keys[i]]:
                if group.getMemberByIdRef(reac) is None:
                    member = group.createMember()
                    member.setIdRef(reac)
        else:
            group = group_list.createGroup()
            group.setName(dbentry["NAME"][0])
            group.setId("G_" + keys[i])  # important for validity (memote/cobra)
            group.setMetaId("meta_" + "G_" + keys[i])
            group.setKind("partonomy")
            group.setSBOTerm("SBO:0000633")  # NAME
            add_cv_term_pathways(keys[i], "KEGG", group)
            for reac in pathway_groups[keys[i]]:
                if group.getMemberByIdRef(reac) is None:
                    member = group.createMember()
                    member.setIdRef(reac)

    return model


def set_kegg_pathways(
    modelpath: str, viaEC: bool = False, viaRC: bool = False
) -> tuple[libModel, list[str]]:
    """Executes all steps to add KEGG pathways as groups

    Args:
        - modelpath (str):
            Path to GEM.

    Returns:
        tuple:
            libSBML model (1) & List of reactions without KEGG Id (2)

            (1) libModel: Modified model with Pathways as groups
            (2) list: Ids of reactions without KEGG annotation
    """

    def _invert_reac_patway_dict(kegg_pathways) -> dict:
        """Group reaction into pathways.

        Args:
            - kegg_pathways (dict):
                Reaction Id as key and KEGG Pathway Id as value. Output of :py:func:`~refinegems.curation.pathways.find_kegg_pathways`.

        Returns:
            dict:
                KEGG Pathway Id as key and reactions Ids as values.
        """
        pathway_groups = {}
        for reaction in kegg_pathways.keys():
            for path in kegg_pathways[reaction]:
                if path not in pathway_groups.keys():
                    pathway_groups[path] = [reaction]
                else:
                    pathway_groups[path].append(reaction)
        return pathway_groups

    # load model with groups enabled
    model = load_model_enable_groups(modelpath)

    # extract information about KEGG and EC numbers from model reactions
    reactions, non_kegg_reactions = _extract_kegg_ec_from_reac(model)

    # @DEBUG .................
    # reactions = {k:reactions[k] for idx,k in enumerate(reactions) if idx < 5}
    # ........................

    # add kegg pathways
    pathways = find_kegg_pathways(reactions, viaEC=viaEC, viaRC=viaRC)
    model_pathways = add_kegg_pathways(model, pathways)

    # add corresponding groups
    pathway_groups = _invert_reac_patway_dict(pathways)
    model_pathway_groups = create_pathway_groups(model_pathways, pathway_groups)

    return model_pathway_groups, non_kegg_reactions


# analyse the pathways in a model
# -------------------------------


def kegg_pathway_analysis(model: cobra.Model) -> KEGGPathwayAnalysisReport:
    """Analyse the pathways that are covered by the model.

    The analysis is based on the KEGG pathway classification and the available
    KEGG pathway identifiers present in the model.

    Note: one reaction can have multiple pathway identifiers associated with it.
    This analysis focuses on the total number of IDs found within the model.

    Args:
        - model (cobra.Model):
            A model loaded with COBRApy.

    Returns:
        KEGGPathwayAnalysisReport:
            The KEGG pathway analysis report.
    """
    # create report
    report = KEGGPathwayAnalysisReport(total_reac=len(model.reactions))

    pathways = dict()
    counter = 0
    # extract KEGG pathway IDs from all reactions
    for r in model.reactions:
        if "kegg.pathway" in r.annotation.keys():
            counter += 1
            anno = r.annotation["kegg.pathway"]
            # case 1: only one annotation found
            if isinstance(anno, str):
                anno = re.sub(r"^[a-z]*", "", anno)
                if anno in pathways:
                    pathways[anno] += 1
                else:
                    pathways[anno] = 1
            # case 2: multiple annotations for one reaction found
            else:
                for x in anno:
                    x = re.sub(r"^[a-z]*", "", x)
                    if x in pathways:
                        pathways[x] += 1
                    else:
                        pathways[x] = 1

    # add counter to report
    report.kegg_count = counter

    # identify global and overview pathway identifier
    global_map = {}
    over_map = {}
    rest = {}
    for k, v in pathways.items():
        if k.startswith("011"):
            global_map[k] = v
        elif k.startswith("012"):
            over_map[k] = v
        else:
            rest[k] = v

    # add IDs in corresponding class to report
    report.kegg_global = global_map
    report.kegg_over = over_map
    report.kegg_paths = rest

    return report



#!/usr/bin/env python
"""Provides functions to compare and visualize multiple models

Can mainly be used to compare growth behaviour of multiple models.
All other stats are shown in the memote report.
"""

__author__ = "Famke Baeuerle, Gwendolyn O. Döbel and Carolin Brune"

################################################################################
# requirements
################################################################################

import matplotlib
import matplotlib.colors

from cobra import Model as cobraModel
from libsbml import Model as libModel
from venn import venn

from ..classes.reports import MultiSBOTermReport, SBOTermReport

################################################################################
# functions
################################################################################


def sbo_terms(models: list[libModel]) -> MultiSBOTermReport:
    """Analyse and compare the SBO term annotations of a given list
    of models.

    Args:
        - models (list[libModel]):
            A list containing models loaded with libSBML.

    Returns:
        MultiSBOTermReport:
            A :py:class:`~refinegems.classes.reports.MultiSBOTermReport` instance.
    """

    sboanalyses = []
    for m in models:
        sboanalyses.append(SBOTermReport(m))

    return MultiSBOTermReport(sboanalyses)


def plot_venn(
    models: list[cobraModel], entity: str, perc: bool = False, rename=None
) -> matplotlib.axes.Axes:
    """Creates Venn diagram to show the overlap of model entities

    Args:
        - models (list[cobraModel]):
            Models loaded with cobrapy
        - entity (str):
            Compare on metabolite|reaction
        - perc (bool, optional):
            True if percentages should be used.
            Defaults to False.
        - rename (dict, optional):
            Rename model ids to custom names.
            Defaults to None.

    Returns:
        matplotlib.axes.Axes:
            Venn diagram
    """
    intersec = {}
    for model in models:
        reas = []
        if entity == "metabolite":
            for rea in model.metabolites:
                reas.append(rea.id)
        if entity == "reaction":
            for rea in model.reactions:
                reas.append(rea.id)
        if rename is not None:
            intersec[rename[model.id]] = set(reas)
        else:
            intersec[model.id] = set(reas)
    if perc:
        fig = venn(intersec, fmt="{percentage:.1f}%")
    else:
        fig = venn(intersec)
    return fig
